import glob
import os
import time
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import Any, List, Tuple, Union

import numpy as np

import vedo.vtkclasses as vtki  # a wrapper for lazy imports

import vedo
from vedo import settings
from vedo import colors
from vedo import utils
from vedo.assembly import Assembly
from vedo.image import Image
from vedo.pointcloud import Points
from vedo.mesh import Mesh
from vedo.volume import Volume

__docformat__ = "google"

__doc__ = """
Submodule to read/write meshes and other objects in different formats,
and other I/O functionalities.
"""

__all__ = [
    "load",
    "read",
    "download",
    "gunzip",
    "loadStructuredPoints",
    "loadStructuredGrid",
    "write",
    "save",
    "export_window",
    "import_window",
    "load_obj",
    "screenshot",
    "ask",
    "Video",
]


# example web page for X3D
_x3d_html_template = """
<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="UTF-8">
  <meta name="viewport" content="width=device-width, initial-scale=1.0">
  <title> vedo with x3d </title>

  <!-- THESE ARE THE RELEVANT LINES: -->
  <script src='https://www.x3dom.org/download/x3dom.js'> </script>
  <link rel='stylesheet' type='text/css' href='https://www.x3dom.org/download/x3dom.css'/>

  <style>
     table, td, th { border: 1px solid black; background-color: powderblue;}
     table {width: 70%; border-collapse: collapse;}
     table th {width: 35%;}
  </style>
</head>

<body style="font-family: Verdana">
  <h1>Example html generated by vedo</h1>
  This example loads a 3D scene from file ~fileoutput generated by
  <a href="https://github.com/marcomusy/vedo">vedo</a>
  (see <a href="https://github.com/marcomusy/vedo/tree/master/examples/other/export_x3d.py">export_x3d.py</a>).
  <br><br>


  <!-- THESE ARE THE RELEVANT LINES: -->
  <x3d width='~widthpx' height='~heightpx'>
     <scene>
        <Inline url="~fileoutput"> </Inline>
     </scene>
  </x3d>

  <h3>Nothing shows up above this line?</h3>
  Enable your browser to load local files:
  <br><b>Firefox</b>: type <code>about:config</code> in the URL bar and
  change <code>privacy.file_unique_origin</code> from <code>True</code> to <code>False</code>
  <br><b>Chrome</b>: from terminal type:
  <code>google-chrome --enable-webgl --allow-file-access-from-files</code>
  (see <a href="https://cmatskas.com/interacting-with-local-data-files-using-chrome/">here</a>)

  <br>
  <h3>Controls:</h3>
  <h4><strong>Examine Mode (activate with key 'e'):</strong></h4>
  <table>
     <tbody>
        <tr class="even description">
           <th>Button</th>
           <th>Function</th>
        </tr>
        <tr>
           <td>Left Button / Left Button + Shift</td>
           <td>Rotate</td>
        </tr>
        <tr>
           <td>Mid Button / Left Button + Ctl</td>
           <td>Pan</td>
        </tr>
        <tr>
           <td>Right Button / Wheel / Left Button + Alt</td>
           <td>Zoom</td>
        </tr>
        <tr>
           <td>Left double click</td>
           <td>Set center of rotation</td>
        </tr>
     </tbody>
  </table>
  <h4><strong>Walk Mode (activate with key 'w'):</strong></h4>
  <table>
     <tbody>
        <tr class="even description">
           <th>Button</th>
           <th>Function</th>
        </tr>
        <tr>
           <td>Left Button</td>
           <td>Move forward</td>
        </tr>
        <tr>
           <td>Right Button</td>
           <td>Move backward</td>
        </tr>
     </tbody>
  </table>
  <h4><strong>Fly Mode (activate with key 'f'):</strong></h4>
  <table>
     <tbody>
        <tr class="even description">
           <th>Button</th>
           <th>Function</th>
        </tr>
        <tr>
           <td>Left Button</td>
           <td>Move forward</td>
        </tr>
        <tr>
           <td>Right Button</td>
           <td>Move backward</td>
        </tr>
     </tbody>
  </table>
  <h3>Non-interactive camera movement</h3>
  <table>
     <tbody>
        <tr class="even description">
           <th>Key</th>
           <th>Function</th>
        </tr>
        <tr>
           <td>r</td>
           <td>reset view</td>
        </tr>
        <tr>
           <td>a</td>
           <td>show all</td>
        </tr>
        <tr>
           <td>u</td>
           <td>upright</td>
        </tr>
     </tbody>
  </table>
</body>
</html>
"""

########################################################################
def load(inputobj: Union[list, str, os.PathLike], unpack=True, force=False) -> Any:
    """
    Load any vedo objects from file or from the web.

    The output will depend on the file extension. See examples below.
    Unzip is made on the fly, if file ends with `.gz`.
    Can load an object directly from a URL address.

    Arguments:
        unpack : (bool)
            unpack MultiBlockData into a flat list of objects.
        force : (bool)
            when downloading a file ignore any previous cached downloads and force a new one.

    Example:
        ```python
        from vedo import dataurl, load, show
        # Return a list of 2 meshes
        g = load([dataurl+'250.vtk', dataurl+'270.vtk'])
        show(g)
        # Return a list of meshes by reading all files in a directory
        # (if directory contains DICOM files then a Volume is returned)
        g = load('mydicomdir/')
        show(g)
        ```
    """
    if isinstance(inputobj, list):
        inputobj = [str(f) for f in inputobj]
    else:
        inputobj = str(inputobj)
    
    acts = []
    if utils.is_sequence(inputobj):
        flist = inputobj
    elif isinstance(inputobj, str) and inputobj.startswith("https://"):
        flist = [inputobj]
    else:
        flist = utils.humansort(glob.glob(inputobj))

    for fod in flist:

        if fod.startswith("https://"):
            fod = download(fod, force=force, verbose=False)

        if os.path.isfile(fod):  ### it's a file

            if fod.endswith(".gz"):
                fod = gunzip(fod)

            a = _load_file(fod, unpack)
            acts.append(a)

        elif os.path.isdir(fod):  ### it's a directory or DICOM
            flist = os.listdir(fod)
            if ".dcm" in flist[0]:  ### it's DICOM
                reader = vtki.new("DICOMImageReader")
                reader.SetDirectoryName(fod)
                reader.Update()
                image = reader.GetOutput()
                vol = Volume(image)
                try:
                    vol.metadata["PixelSpacing"] = reader.GetPixelSpacing()
                    vol.metadata["Width"] = reader.GetWidth()
                    vol.metadata["Height"] = reader.GetHeight()
                    vol.metadata["PositionPatient"] = reader.GetImagePositionPatient()
                    vol.metadata["OrientationPatient"] = reader.GetImageOrientationPatient()
                    vol.metadata["BitsAllocated"] = reader.GetBitsAllocated()
                    vol.metadata["PixelRepresentation"] = reader.GetPixelRepresentation()
                    vol.metadata["NumberOfComponents"] = reader.GetNumberOfComponents()
                    vol.metadata["TransferSyntaxUID"] = reader.GetTransferSyntaxUID()
                    vol.metadata["RescaleSlope"] = reader.GetRescaleSlope()
                    vol.metadata["RescaleOffset"] = reader.GetRescaleOffset()
                    vol.metadata["PatientName"] = reader.GetPatientName()
                    vol.metadata["StudyUID"] = reader.GetStudyUID()
                    vol.metadata["StudyID"] = reader.GetStudyID()
                    vol.metadata["GantryAngle"] = reader.GetGantryAngle()
                except Exception as e:
                    vedo.logger.warning(f"Cannot read DICOM metadata: {e}")
                acts.append(vol)

            else:  ### it's a normal directory
                utils.humansort(flist)
                for ifile in flist:
                    a = _load_file(fod + "/" + ifile, unpack)
                    acts.append(a)
        else:
            vedo.logger.error(f"in load(), cannot find {fod}")

    if len(acts) == 1:
        if "numpy" in str(type(acts[0])):
            return acts[0]
        if not acts[0]:
            vedo.logger.error(f"in load(), cannot load {inputobj}")
        return acts[0]

    if len(acts) == 0:
        vedo.logger.error(f"in load(), cannot load {inputobj}")
        return None

    else:
        return acts

########################################################################
def _load_file(filename, unpack):
    fl = filename.lower()

    ########################################################## other formats:
    if fl.endswith(".xml") or fl.endswith(".xml.gz") or fl.endswith(".xdmf"):
        # Fenics tetrahedral file
        objt = loadDolfin(filename)
    elif fl.endswith(".neutral") or fl.endswith(".neu"):  # neutral tets
        objt = loadNeutral(filename)
    elif fl.endswith(".gmsh"):  # gmesh file
        objt = loadGmesh(filename)
    elif fl.endswith(".pcd"):  # PCL point-cloud format
        objt = loadPCD(filename)
        objt.properties.SetPointSize(2)
    elif fl.endswith(".off"):
        objt = loadOFF(filename)
    elif fl.endswith(".3ds"):  # 3ds format
        objt = load3DS(filename)
    elif fl.endswith(".wrl"):
        importer = vtki.new("VRMLImporter")
        importer.SetFileName(filename)
        importer.Read()
        importer.Update()
        actors = importer.GetRenderer().GetActors()  # vtkActorCollection
        actors.InitTraversal()
        wacts = []
        for i in range(actors.GetNumberOfItems()):
            act = actors.GetNextActor()
            m = Mesh(act.GetMapper().GetInput())
            m.actor = act
            wacts.append(m)
        objt = Assembly(wacts)

    ######################################################## volumetric:
    elif (
        fl.endswith(".tif")
        or fl.endswith(".tiff")
        or fl.endswith(".slc")
        or fl.endswith(".vti")
        or fl.endswith(".mhd")
        or fl.endswith(".nrrd")
        or fl.endswith(".nii")
        or fl.endswith(".dem")
    ):
        img = loadImageData(filename)
        objt = Volume(img)

    ######################################################### 2D images:
    elif (
        fl.endswith(".png")
        or fl.endswith(".jpg")
        or fl.endswith(".bmp")
        or fl.endswith(".jpeg")
        or fl.endswith(".gif")
    ):
        if ".png" in fl:
            picr = vtki.new("PNGReader")
        elif ".jpg" in fl or ".jpeg" in fl:
            picr = vtki.new("JPEGReader")
        elif ".bmp" in fl:
            picr = vtki.new("BMPReader")
        elif ".gif" in fl:
            from PIL import Image as PILImage, ImageSequence

            img = PILImage.open(filename)
            frames = []
            for frame in ImageSequence.Iterator(img):
                a = np.array(frame.convert("RGB").getdata(), dtype=np.uint8)
                a = a.reshape([frame.size[1], frame.size[0], 3])
                frames.append(Image(a))
            return frames

        picr.SetFileName(filename)
        picr.Update()
        objt = Image(picr.GetOutput())

    ######################################################### multiblock:
    elif fl.endswith(".vtm") or fl.endswith(".vtmb"):
        read = vtki.new("XMLMultiBlockDataReader")
        read.SetFileName(filename)
        read.Update()
        mb = read.GetOutput()
        if unpack:
            acts = []
            for i in range(mb.GetNumberOfBlocks()):
                b = mb.GetBlock(i)
                if isinstance(
                    b,
                    (
                        vtki.vtkPolyData,
                        vtki.vtkStructuredGrid,
                        vtki.vtkRectilinearGrid,
                    ),
                ):
                    acts.append(Mesh(b))
                elif isinstance(b, vtki.vtkImageData):
                    acts.append(Volume(b))
                elif isinstance(b, vtki.vtkUnstructuredGrid):
                    acts.append(vedo.UnstructuredGrid(b))
            return acts
        return mb
    
    ######################################################### assembly:
    elif fl.endswith(".npy"):
        data = np.load(filename, allow_pickle=True)
        try:
            # old format with a single object
            meshs = [from_numpy(dd) for dd in data]
        except TypeError:
            data = data.item()
            meshs = []
            for ad in data["objects"][0]["parts"]:
                obb = from_numpy(ad)
                meshs.append(obb)
        return Assembly(meshs)

    ###########################################################
    elif fl.endswith(".geojson"):
        return loadGeoJSON(filename)

    elif fl.endswith(".pvd"):
        return loadPVD(filename)

    ########################################################### polygonal mesh:
    else:
        if fl.endswith(".vtk"):  # read all legacy vtk types
            reader = vtki.new("DataSetReader")
            reader.ReadAllScalarsOn()
            reader.ReadAllVectorsOn()
            reader.ReadAllTensorsOn()
            reader.ReadAllFieldsOn()
            reader.ReadAllNormalsOn()
            reader.ReadAllColorScalarsOn()
        elif fl.endswith(".ply"):
            reader = vtki.new("PLYReader")
        elif fl.endswith(".obj"):
            reader = vtki.new("OBJReader")
            reader.SetGlobalWarningDisplay(0) # suppress warnings issue #980
        elif fl.endswith(".stl"):
            reader = vtki.new("STLReader")
        elif fl.endswith(".byu") or fl.endswith(".g"):
            reader = vtki.new("BYUReader")
        elif fl.endswith(".foam"):  # OpenFoam
            reader = vtki.new("OpenFOAMReader")
        elif fl.endswith(".pvd"):
            reader = vtki.new("XMLGenericDataObjectReader")
        elif fl.endswith(".vtp"):
            reader = vtki.new("XMLPolyDataReader")
        elif fl.endswith(".vts"):
            reader = vtki.new("XMLStructuredGridReader")
        elif fl.endswith(".vtu"):
            reader = vtki.new("XMLUnstructuredGridReader")
        elif fl.endswith(".vtr"):
            reader = vtki.new("XMLRectilinearGridReader")
        elif fl.endswith(".pvtr"):
            reader = vtki.new("XMLPRectilinearGridReader")
        elif fl.endswith("pvtu"):
            reader = vtki.new("XMLPUnstructuredGridReader")
        elif fl.endswith(".txt") or fl.endswith(".xyz") or fl.endswith(".dat"):
            reader = vtki.new("ParticleReader")  # (format is x, y, z, scalar)
        elif fl.endswith(".facet"):
            reader = vtki.new("FacetReader")
        else:
            return None

        reader.SetFileName(filename)
        reader.Update()
        routput = reader.GetOutput()

        if not routput:
            vedo.logger.error(f"unable to load {filename}") 
            return None

        if isinstance(routput, vtki.vtkUnstructuredGrid):
            objt = vedo.UnstructuredGrid(routput)

        else:
            objt = Mesh(routput)
            if fl.endswith(".txt") or fl.endswith(".xyz") or fl.endswith(".dat"):
                objt.point_size(4)

    objt.filename = filename
    objt.file_size, objt.created = file_info(filename)
    return objt


def download(url: str, force=False, verbose=True) -> str:
    """
    Retrieve a file from a URL, save it locally and return its path.
    Use `force=True` to force a reload and discard cached copies.
    """
    if not url.startswith("https://"):
        # assume it's a file so no need to download
        return url
    url = url.replace("www.dropbox", "dl.dropbox")

    if "github.com" in url:
        url = url.replace("/blob/", "/raw/")

    basename = os.path.basename(url)

    if "?" in basename:
        basename = basename.split("?")[0]

    home_directory = os.path.expanduser("~")
    cachedir = os.path.join(home_directory, settings.cache_directory, "vedo")
    fname = os.path.join(cachedir, basename)
    # Create the directory if it does not exist
    if not os.path.exists(cachedir):
        os.makedirs(cachedir)

    if not force and os.path.exists(fname):
        if verbose:
            colors.printc("reusing cached file:", fname)
        return fname

    try:
        from urllib.request import urlopen, Request
        req = Request(url, headers={"User-Agent": "Mozilla/5.0"})
        if verbose:
            colors.printc("reading", basename, "from", url.split("/")[2][:40], "...", end="")

    except ImportError:
        import urllib2
        import contextlib
        urlopen = lambda url_: contextlib.closing(urllib2.urlopen(url_))
        req = url
        if verbose:
            colors.printc("reading", basename, "from", url.split("/")[2][:40], "...", end="")

    with urlopen(req) as response, open(fname, "wb") as output:
        output.write(response.read())

    if verbose:
        colors.printc(" done.")
    return fname


########################################################################
# def download_new(url, to_local_file="", force=False, verbose=True):
#     """
#     Downloads a file from `url` to `to_local_file` if the local copy is outdated.

#     Arguments:
#         url : (str)
#             The URL to download the file from.
#         to_local_file : (str)
#             The local file name to save the file to. 
#             If not specified, the file name will be the same as the remote file name
#             in the directory specified by `settings.cache_directory + "/vedo"`.
#         force : (bool)
#             Force a new download even if the local file is up to date.
#         verbose : (bool)
#             Print verbose messages.
#     """
#     if not url.startswith("https://"):
#         if os.path.exists(url):
#             # Assume the url is already the local file path
#             return url
#         else:
#             raise FileNotFoundError(f"File not found: {url}")

#     from datetime import datetime
#     import requests

#     url = url.replace("www.dropbox", "dl.dropbox")

#     if "github.com" in url:
#         url = url.replace("/blob/", "/raw/")

#     # Get the user's home directory
#     home_directory = os.path.expanduser("~")

#     # Define the path for the cache directory
#     cachedir = os.path.join(home_directory, settings.cache_directory, "vedo")

#     # Create the directory if it does not exist
#     if not os.path.exists(cachedir):
#         os.makedirs(cachedir)

#     if not to_local_file:
#         basename = os.path.basename(url)
#         if "?" in basename:
#             basename = basename.split("?")[0]
#         to_local_file = os.path.join(cachedir, basename)
#         if verbose: print(f"Using local file name: {to_local_file}")

#     # Check if the local file exists and get its last modified time
#     if os.path.exists(to_local_file):
#         to_local_file_modified_time = os.path.getmtime(to_local_file)
#     else:
#         to_local_file_modified_time = 0

#     # Send a HEAD request to get last modified time of the remote file
#     response = requests.head(url)
#     if 'Last-Modified' in response.headers:
#         remote_file_modified_time = datetime.strptime(
#             response.headers['Last-Modified'], '%a, %d %b %Y %H:%M:%S GMT'
#         ).timestamp()
#     else:
#         # If the Last-Modified header not available, assume file needs to be downloaded
#         remote_file_modified_time = float('inf')

#     # Download the file if the remote file is newer
#     if force or remote_file_modified_time > to_local_file_modified_time:
#         response = requests.get(url)
#         with open(to_local_file, 'wb') as file:
#             file.write(response.content)
#             if verbose: print(f"Downloaded file from {url} -> {to_local_file}")
#     else:
#         if verbose: print("Local file is up to date.")
#     return to_local_file


########################################################################
def gunzip(filename: str) -> str:
    """Unzip a `.gz` file to a temporary file and returns its path."""
    if not filename.endswith(".gz"):
        # colors.printc("gunzip() error: file must end with .gz", c='r')
        return filename

    import gzip

    tmp_file = NamedTemporaryFile(delete=False)
    tmp_file.name = os.path.join(
        os.path.dirname(tmp_file.name), os.path.basename(filename).replace(".gz", "")
    )
    inF = gzip.open(filename, "rb")
    with open(tmp_file.name, "wb") as outF:
        outF.write(inF.read())
    inF.close()
    return tmp_file.name

########################################################################
def file_info(file_path: str) -> Tuple[str, str]:
    """Return the file size and creation time of input file"""
    siz, created = "", ""
    if os.path.isfile(file_path):
        f_info = os.stat(file_path)
        num = f_info.st_size
        for x in ["B", "KB", "MB", "GB", "TB"]:
            if num < 1024.0:
                break
            num /= 1024.0
        siz = "%3.1f%s" % (num, x)
        created = time.ctime(os.path.getmtime(file_path))
    return siz, created


###################################################################
def loadStructuredPoints(filename: Union[str, os.PathLike], as_points=True):
    """
    Load and return a `vtkStructuredPoints` object from file.
    
    If `as_points` is True, return a `Points` object
    instead of a `vtkStructuredPoints`.
    """
    filename = str(filename)
    reader = vtki.new("StructuredPointsReader")
    reader.SetFileName(filename)
    reader.Update()
    if as_points:
        v2p = vtki.new("ImageToPoints")
        v2p.SetInputData(reader.GetOutput())
        v2p.Update()
        pts = Points(v2p.GetOutput())
        return pts
    return reader.GetOutput()

########################################################################
def loadStructuredGrid(filename: Union[str, os.PathLike]):
    """Load and return a `vtkStructuredGrid` object from file."""
    filename = str(filename)
    if filename.endswith(".vts"):
        reader = vtki.new("XMLStructuredGridReader")
    else:
        reader = vtki.new("StructuredGridReader")
    reader.SetFileName(filename)
    reader.Update()
    return reader.GetOutput()


###################################################################
def load3DS(filename: Union[str, os.PathLike]) -> Assembly:
    """Load `3DS` file format from file."""
    filename = str(filename)
    renderer = vtki.vtkRenderer()
    renWin = vtki.vtkRenderWindow()
    renWin.AddRenderer(renderer)

    importer = vtki.new("3DSImporter")
    importer.SetFileName(filename)
    importer.ComputeNormalsOn()
    importer.SetRenderWindow(renWin)
    importer.Update()

    actors = renderer.GetActors()  # vtkActorCollection
    acts = []
    for i in range(actors.GetNumberOfItems()):
        a = actors.GetItemAsObject(i)
        acts.append(a)
    del renWin

    wrapped_acts = []
    for a in acts:
        try:
            newa = Mesh(a.GetMapper().GetInput())
            newa.actor = a
            wrapped_acts.append(newa)
            # print("loaded 3DS object", [a])
        except:
            print("ERROR: cannot load 3DS object part", [a])
    return vedo.Assembly(wrapped_acts)

########################################################################
def loadOFF(filename: Union[str, os.PathLike]) -> Mesh:
    """Read the OFF file format (polygonal mesh)."""
    filename = str(filename)
    with open(filename, "r", encoding="UTF-8") as f:
        lines = f.readlines()

    vertices = []
    faces = []
    NumberOfVertices = 0
    i = -1
    for text in lines:
        if len(text) == 0:
            continue
        if text == "\n":
            continue
        if "#" in text:
            continue
        if "OFF" in text:
            continue

        ts = text.split()
        n = len(ts)

        if not NumberOfVertices and n > 1:
            NumberOfVertices, NumberOfFaces = int(ts[0]), int(ts[1])
            continue
        i += 1

        if i < NumberOfVertices and n == 3:
            x, y, z = float(ts[0]), float(ts[1]), float(ts[2])
            vertices.append([x, y, z])

        ids = []
        if NumberOfVertices <= i < (NumberOfVertices + NumberOfFaces + 1) and n > 2:
            ids += [int(xx) for xx in ts[1:]]
            faces.append(ids)

    return Mesh(utils.buildPolyData(vertices, faces))

########################################################################
def loadGeoJSON(filename: Union[str, os.PathLike]) -> Mesh:
    """Load GeoJSON files."""
    filename = str(filename)
    jr = vtki.new("GeoJSONReader")
    jr.SetFileName(filename)
    jr.Update()
    return Mesh(jr.GetOutput())

########################################################################
def loadDolfin(filename: Union[str, os.PathLike]) -> Union[Mesh, "vedo.TetMesh", None]:
    """
    Reads a `Fenics/Dolfin` file format (.xml or .xdmf).

    Return a `Mesh` or a `TetMesh` object.
    """
    filename = str(filename)
    try:
        import dolfin
    except ImportError:
        vedo.logger.error("loadDolfin(): dolfin module not found. Install with:")
        vedo.logger.error("  conda create -n fenics -c conda-forge fenics")
        vedo.logger.error("  conda install conda-forge::mshr")
        vedo.logger.error("  conda activate fenics")
        return None

    if filename.lower().endswith(".xdmf"):
        f = dolfin.XDMFFile(filename)
        m = dolfin.Mesh()
        f.read(m)
    else:
        m = dolfin.Mesh(filename)
    
    cells = m.cells()
    verts = m.coordinates()

    if cells.size and verts.size:
        if len(cells[0]) == 4:  # tetrahedral mesh
            return vedo.TetMesh([verts, cells])
        elif len(cells[0]) == 3:  # triangular mesh
            return Mesh([verts, cells])
    
    return None


########################################################################
def loadPVD(filename: Union[str, os.PathLike]) -> Union[List[Any], None]:
    """Read paraview files."""
    filename = str(filename)
    import xml.etree.ElementTree as et

    tree = et.parse(filename)

    dname = os.path.dirname(filename)
    if not dname:
        dname = "."

    listofobjs = []
    for coll in tree.getroot():
        for dataset in coll:
            fname = dataset.get("file")
            if not fname:
                continue
            ob = load(dname + "/" + fname)
            tm = dataset.get("timestep")
            if tm:
                ob.time = tm
            listofobjs.append(ob)
    if len(listofobjs) == 1:
        return listofobjs[0]
    if len(listofobjs) == 0:
        return None
    return listofobjs

########################################################################
def loadNeutral(filename: Union[str, os.PathLike]) -> "vedo.TetMesh":
    """
    Reads a `Neutral` tetrahedral file format.

    Returns an `TetMesh` object.
    """
    filename = str(filename)
    with open(filename, "r", encoding="UTF-8") as f:
        lines = f.readlines()

    ncoords = int(lines[0])
    coords = []
    for i in range(1, ncoords + 1):
        x, y, z = lines[i].split()
        coords.append([float(x), float(y), float(z)])

    ntets = int(lines[ncoords + 1])
    idolf_tets = []
    for i in range(ncoords + 2, ncoords + ntets + 2):
        text = lines[i].split()
        v0, v1, v2, v3 = int(text[1])-1, int(text[2])-1, int(text[3])-1, int(text[4])-1
        idolf_tets.append([v0, v1, v2, v3])

    return vedo.TetMesh([coords, idolf_tets])

########################################################################
def loadGmesh(filename: Union[str, os.PathLike]) -> Mesh:
    """Reads a `gmesh` file format. Return an `Mesh` object."""
    filename = str(filename)
    with open(filename, "r", encoding="UTF-8") as f:
        lines = f.readlines()

    nnodes = 0
    index_nodes = 0
    for i, line in enumerate(lines):
        if "$Nodes" in line:
            index_nodes = i + 1
            nnodes = int(lines[index_nodes])
            break
    node_coords = []
    for i in range(index_nodes + 1, index_nodes + 1 + nnodes):
        cn = lines[i].split()
        node_coords.append([float(cn[1]), float(cn[2]), float(cn[3])])

    nelements = 0
    index_elements = 0
    for i, line in enumerate(lines):
        if "$Elements" in line:
            index_elements = i + 1
            nelements = int(lines[index_elements])
            break
    elements = []
    for i in range(index_elements + 1, index_elements + 1 + nelements):
        ele = lines[i].split()
        elements.append([int(ele[-3]), int(ele[-2]), int(ele[-1])])

    poly = utils.buildPolyData(node_coords, elements, index_offset=1)
    return Mesh(poly)

########################################################################
def loadPCD(filename: Union[str, os.PathLike]) -> Points:
    """Return a `Mesh` made of only vertex points
    from the `PointCloud` library file format.
    
    Returns an `Points` object.
    """
    filename = str(filename)
    with open(filename, "r", encoding="UTF-8") as f:
        lines = f.readlines()

    start = False
    pts = []
    N, expN = 0, 0
    for text in lines:
        if start:
            if N >= expN:
                break
            l = text.split()
            pts.append([float(l[0]), float(l[1]), float(l[2])])
            N += 1
        if not start and "POINTS" in text:
            expN = int(text.split()[1])
        if not start and "DATA ascii" in text:
            start = True
    if expN != N:
        vedo.logger.warning(f"Mismatch in PCD file {expN} != {len(pts)}")
    poly = utils.buildPolyData(pts)
    return Points(poly).point_size(4)

#########################################################################
def from_numpy(d: dict) -> Mesh:
    # recreate a mesh from numpy arrays
    keys = d.keys()

    points = d["points"]
    cells = d["cells"] if "cells" in keys else None
    lines = d["lines"] if "lines" in keys else None

    msh = Mesh([points, cells, lines])

    if "pointdata" in keys and isinstance(d["pointdata"], dict):
        for arrname, arr in d["pointdata"].items():
            msh.pointdata[arrname] = arr
    if "celldata" in keys and isinstance(d["celldata"], dict):
        for arrname, arr in d["celldata"].items():
            msh.celldata[arrname] = arr
    if "metadata" in keys and isinstance(d["metadata"], dict):
        for arrname, arr in d["metadata"].items():
            msh.metadata[arrname] = arr

    prp = msh.properties
    prp.SetAmbient(d['ambient'])
    prp.SetDiffuse(d['diffuse'])
    prp.SetSpecular(d['specular'])
    prp.SetSpecularPower(d['specularpower'])
    prp.SetSpecularColor(d['specularcolor'])

    prp.SetInterpolation(0)
    # prp.SetInterpolation(d['shading'])

    prp.SetOpacity(d['alpha'])
    prp.SetRepresentation(d['representation'])
    prp.SetPointSize(d['pointsize'])
    if d['color'] is not None:
        msh.color(d['color'])
    if "lighting_is_on" in d.keys(): 
        prp.SetLighting(d['lighting_is_on'])
    # Must check keys for backwards compatibility:
    if "linecolor" in d.keys() and d['linecolor'] is not None:
        msh.linecolor(d['linecolor'])
    if "backcolor" in d.keys() and d['backcolor'] is not None:
        msh.backcolor(d['backcolor'])

    if d['linewidth'] is not None:
        msh.linewidth(d['linewidth'])
    if "edge_visibility" in d.keys(): 
        prp.SetEdgeVisibility(d['edge_visibility']) # new

    lut_list  = d["LUT"]
    lut_range = d["LUT_range"]
    ncols = len(lut_list)
    lut = vtki.vtkLookupTable()
    lut.SetNumberOfTableValues(ncols)
    lut.SetRange(lut_range)
    for i in range(ncols):
        r, g, b, a = lut_list[i]
        lut.SetTableValue(i, r, g, b, a)
    lut.Build()
    msh.mapper.SetLookupTable(lut)
    msh.mapper.SetScalarRange(lut_range)

    try: # NEW in vedo 5.0
        arname = d["array_name_to_color_by"]
        msh.mapper.SetArrayName(arname)
        msh.mapper.SetInterpolateScalarsBeforeMapping(
            d["interpolate_scalars_before_mapping"])
        msh.mapper.SetUseLookupTableScalarRange(
            d["use_lookup_table_scalar_range"])
        msh.mapper.SetScalarRange(d["scalar_range"])
        msh.mapper.SetScalarVisibility(d["scalar_visibility"])
        msh.mapper.SetScalarMode(d["scalar_mode"])
        msh.mapper.SetColorMode(d["color_mode"])
        if d["scalar_visibility"]:
            if d["scalar_mode"] == 1:
                msh.dataset.GetPointData().SetActiveScalars(arname)
            if d["scalar_mode"] == 2:
                msh.dataset.GetCellData().SetActiveScalars(arname)

        if "texture_array" in keys and d["texture_array"] is not None:
            # recreate a vtkTexture object from numpy arrays:
            t = vtki.vtkTexture()
            t.SetInterpolate(d["texture_interpolate"])
            t.SetRepeat(d["texture_repeat"])
            t.SetQuality(d["texture_quality"])
            t.SetColorMode(d["texture_color_mode"])
            t.SetMipmap(d["texture_mipmap"])
            t.SetBlendingMode(d["texture_blending_mode"])
            t.SetEdgeClamp(d["texture_edge_clamp"])
            t.SetBorderColor(d["texture_border_color"])
            msh.actor.SetTexture(t)
            tcarray = None
            for arrname in msh.pointdata.keys():
                if "Texture" in arrname or "TCoord" in arrname:
                    tcarray = arrname
                    break
            if tcarray is not None:
                t.SetInputData(vedo.Image(d["texture_array"]).dataset)
                msh.pointdata.select_texture_coords(tcarray)

        # print("color_mode", d["color_mode"])
        # print("scalar_mode", d["scalar_mode"])
        # print("scalar_range", d["scalar_range"])
        # print("scalar_visibility", d["scalar_visibility"])
        # print("array_name_to_color_by", arname)
    except KeyError:
        pass

    if "time" in keys: msh.time = d["time"]
    if "name" in keys: msh.name = d["name"]
    # if "info" in keys: msh.info = d["info"]
    if "filename" in keys: msh.filename = d["filename"]
    if "pickable" in keys: msh.pickable(d["pickable"])
    if "dragable" in keys: msh.draggable(d["dragable"])
    return msh

#############################################################################
def _import_npy(fileinput: Union[str, os.PathLike]) -> "vedo.Plotter":
    """Import a vedo scene from numpy format."""
    fileinput = str(fileinput)

    fileinput = download(fileinput, verbose=False, force=True)
    if fileinput.endswith(".npy"):
        data = np.load(fileinput, allow_pickle=True, encoding="latin1").flatten()[0]
    elif fileinput.endswith(".npz"):
        data = np.load(fileinput, allow_pickle=True)["vedo_scenes"][0]

    if "use_parallel_projection" in data.keys():
        vedo.settings.use_parallel_projection = data["use_parallel_projection"]
    if "use_polygon_offset" in data.keys():
        vedo.settings.use_polygon_offset = data["use_polygon_offset"]
    if "polygon_offset_factor" in data.keys():
        vedo.settings.polygon_offset_factor = data["polygon_offset_factor"]
    if "polygon_offset_units" in data.keys():
        vedo.settings.polygon_offset_units = data["polygon_offset_units"]
    if "interpolate_scalars_before_mapping" in data.keys():
        vedo.settings.interpolate_scalars_before_mapping = data["interpolate_scalars_before_mapping"]
    if "default_font" in data.keys():
        vedo.settings.default_font = data["default_font"]
    if "use_depth_peeling" in data.keys():
        vedo.settings.use_depth_peeling = data["use_depth_peeling"]

    axes = data.pop("axes", 4) # UNUSED
    title = data.pop("title", "")
    backgrcol  = data.pop("backgrcol", "white")
    backgrcol2 = data.pop("backgrcol2", None)
    cam = data.pop("camera", None)

    if data["shape"] != (1, 1):
        data["size"] = "auto"  # disable size

    plt = vedo.Plotter(
        size=data["size"],  # not necessarily a good idea to set it
        axes=axes,          # must be zero to avoid recreating the axes
        title=title,
        bg=backgrcol,
        bg2=backgrcol2,
    )

    if cam:
        if "pos" in cam.keys():
            plt.camera.SetPosition(cam["pos"])
        if "focalPoint" in cam.keys(): # obsolete
            plt.camera.SetFocalPoint(cam["focalPoint"])
        if "focal_point" in cam.keys():
            plt.camera.SetFocalPoint(cam["focal_point"])
        if "viewup" in cam.keys():
            plt.camera.SetViewUp(cam["viewup"])
        if "distance" in cam.keys():
            plt.camera.SetDistance(cam["distance"])
        if "clippingRange" in cam.keys(): # obsolete
            plt.camera.SetClippingRange(cam["clippingRange"])
        if "clipping_range" in cam.keys():
            plt.camera.SetClippingRange(cam["clipping_range"])
        if "parallel_scale" in cam.keys():
            plt.camera.SetParallelScale(cam["parallel_scale"])

    ##############################################
    objs = []
    for d in data["objects"]:
        ### Mesh
        if d['type'].lower() == 'mesh':
            obj = from_numpy(d)

        ### Assembly
        elif d['type'].lower() == 'assembly':
            assacts = []
            for ad in d["actors"]:
                assacts.append(from_numpy(ad))
            obj = Assembly(assacts)
            obj.SetScale(d["scale"])
            obj.SetPosition(d["position"])
            obj.SetOrientation(d["orientation"])
            obj.SetOrigin(d["origin"])

        ### Volume
        elif d['type'].lower() == 'volume':
            obj = Volume(d["array"])
            obj.spacing(d["spacing"])
            obj.origin(d["origin"])
            if "jittering" in d.keys(): obj.jittering(d["jittering"])
            obj.mode(d["mode"])
            obj.color(d["color"])
            obj.alpha(d["alpha"])
            obj.alpha_gradient(d["alphagrad"])

        ### TetMesh
        elif d['type'].lower() == 'tetmesh':
            raise NotImplementedError("TetMesh not supported yet")

        ### ScalarBar2D
        elif d['type'].lower() == 'scalarbar2d':
            raise NotImplementedError("ScalarBar2D not supported yet")

        ### Image
        elif d['type'].lower() == 'image':
            obj = Image(d["array"])
            obj.alpha(d["alpha"])
            obj.actor.SetScale(d["scale"])
            obj.actor.SetPosition(d["position"])
            obj.actor.SetOrientation(d["orientation"])
            obj.actor.SetOrigin(d["origin"])

        ### Text2D
        elif d['type'].lower() == 'text2d':
            obj = vedo.shapes.Text2D(d["text"], font=d["font"], c=d["color"])
            obj.pos(d["position"]).size(d["size"])
            obj.background(d["bgcol"], d["alpha"])
            if d["frame"]:
                obj.frame(d["bgcol"])
        
        else:
            obj = None
            # vedo.logger.warning(f"Cannot import object {d}")
            pass

        if obj:
            keys = d.keys()
            if "time" in keys: obj.time = d["time"]
            if "name" in keys: obj.name = d["name"]
            # if "info" in keys: obj.info = d["info"]
            if "filename" in keys: obj.filename = d["filename"]
            objs.append(obj)

    plt.add(objs)
    plt.resetcam = False
    return plt

###########################################################
def loadImageData(filename: Union[str, os.PathLike]) -> Union[vtki.vtkImageData, None]:
    """Read and return a `vtkImageData` object from file."""
    filename = str(filename)
    if ".tif" in filename.lower():
        reader = vtki.new("TIFFReader")
        # print("GetOrientationType ", reader.GetOrientationType())
        reader.SetOrientationType(vedo.settings.tiff_orientation_type)
    elif ".slc" in filename.lower():
        reader = vtki.new("SLCReader")
        if not reader.CanReadFile(filename):
            vedo.logger.error(f"sorry, bad SLC file {filename}")
            return None
    elif ".vti" in filename.lower():
        reader = vtki.new("XMLImageDataReader")
    elif ".mhd" in filename.lower():
        reader = vtki.new("MetaImageReader")
    elif ".dem" in filename.lower():
        reader = vtki.new("DEMReader")
    elif ".nii" in filename.lower():
        reader = vtki.new("NIFTIImageReader")
    elif ".nrrd" in filename.lower():
        reader = vtki.new("NrrdReader")
        if not reader.CanReadFile(filename):
            vedo.logger.error(f"sorry, bad NRRD file {filename}")
            return None
    else:
        vedo.logger.error(f"cannot read file {filename}")
        return None
    reader.SetFileName(filename)
    reader.Update()
    return reader.GetOutput()

###########################################################
def write(objct: Any, fileoutput: Union[str, os.PathLike], binary=True) -> Any:
    """
    Write object to file. Same as `save()`.

    Supported extensions are:
    
    - `vtk, vti, ply, obj, stl, byu, vtp, vti, mhd, xyz, xml, tif, png, bmp`
    """
    fileoutput = str(fileoutput)

    ###############################
    if isinstance(objct, Assembly):
        dd = to_numpy(objct)
        sdict = {"objects": [dd]}
        np.save(fileoutput, sdict)
        return objct
    
    ###############################
    obj = objct.dataset

    try:
        # check if obj is a Mesh.actor and has a transform
        M = objct.actor.GetMatrix()
        if M and not M.IsIdentity():
            obj = objct.apply_transform_from_actor()
            obj = objct.dataset
            vedo.logger.info(
                f"object '{objct.name}' "
                "was manually moved. Writing uses current position."
            )
    except:
        pass

    fr = fileoutput.lower()
    if fr.endswith(".vtk"):
        writer = vtki.new("DataSetWriter")
    elif fr.endswith(".ply"):
        writer = vtki.new("PLYWriter")
        writer.AddComment("PLY file generated by vedo")
        lut = objct.mapper.GetLookupTable()
        if lut:
            pscal = obj.GetPointData().GetScalars()
            if not pscal:
                pscal = obj.GetCellData().GetScalars()
            if pscal and pscal.GetName():
                writer.SetArrayName(pscal.GetName())
            writer.SetLookupTable(lut)
    elif fr.endswith(".stl"):
        writer = vtki.new("STLWriter")
    elif fr.endswith(".vtp"):
        writer = vtki.new("XMLPolyDataWriter")
    elif fr.endswith(".vtu"):
        writer = vtki.new("XMLUnstructuredGridWriter")
    elif fr.endswith(".xyz"):
        writer = vtki.new("SimplePointsWriter")
    elif fr.endswith(".facet"):
        writer = vtki.new("FacetWriter")
    elif fr.endswith(".vti"):
        writer = vtki.new("XMLImageDataWriter")
    elif fr.endswith(".vtr"):
        writer = vtki.new("XMLRectilinearGridWriter")
    elif fr.endswith(".vtm"):
        g = vtki.new("MultiBlockDataGroupFilter")
        for ob in objct:
            try:
                g.AddInputData(ob)
            except TypeError:
                vedo.logger.warning("cannot save object of type", type(ob))
        g.Update()
        mb = g.GetOutputDataObject(0)
        wri = vtki.new("vtkXMLMultiBlockDataWriter")
        wri.SetInputData(mb)
        wri.SetFileName(fileoutput)
        wri.Write()
        return objct
    elif fr.endswith(".mhd"):
        writer = vtki.new("MetaImageWriter")
    elif fr.endswith(".nii"):
        writer = vtki.new("NIFTIImageWriter")
    elif fr.endswith(".png"):
        writer = vtki.new("PNGWriter")
    elif fr.endswith(".jpg"):
        writer = vtki.new("JPEGWriter")
    elif fr.endswith(".bmp"):
        writer = vtki.new("BMPWriter")
    elif fr.endswith(".tif") or fr.endswith(".tiff"):
        writer = vtki.new("TIFFWriter")
        writer.SetFileDimensionality(len(obj.GetDimensions()))
    elif fr.endswith(".obj"):
        with open(fileoutput, "w", encoding="UTF-8") as outF:
            outF.write("# OBJ file format with ext .obj\n")
            outF.write("# File generated by vedo\n")

            for p in objct.vertices:
                outF.write("v {:.8g} {:.8g} {:.8g}\n".format(*p))

            ptxt = objct.dataset.GetPointData().GetTCoords()
            if ptxt:
                ntxt = utils.vtk2numpy(ptxt)
                for vt in ntxt:
                    outF.write("vt " + str(vt[0]) + " " + str(vt[1]) + " 0.0\n")

            if isinstance(objct, Mesh):
                for i, f in enumerate(objct.cells):
                    fs = ""
                    for fi in f:
                        if ptxt:
                            fs += f" {fi+1}/{fi+1}"
                        else:
                            fs += f" {fi+1}"
                    outF.write(f"f{fs}\n")

                for l in objct.lines:
                    ls = ""
                    for li in l:
                        ls += str(li + 1) + " "
                    outF.write(f"l {ls}\n")
        return objct

    elif fr.endswith(".off"):
        with open(fileoutput, "w", encoding="UTF-8") as outF:
            outF.write("OFF\n")
            outF.write(str(objct.npoints) + " " + str(objct.ncells) + " 0\n\n")
            for p in objct.vertices:
                outF.write(" ".join([str(i) for i in p]) + "\n")
            for c in objct.cells:
                outF.write(str(len(c)) + " " + " ".join([str(i) for i in c]) + "\n")
        return objct

    elif fr.endswith(".xml"):  # write tetrahedral dolfin xml
        vertices = objct.vertices.astype(str)
        faces = np.array(objct.cells).astype(str)
        ncoords = vertices.shape[0]
        with open(fileoutput, "w", encoding="UTF-8") as outF:
            outF.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            outF.write('<dolfin xmlns:dolfin="http://www.fenicsproject.org">\n')

            if len(faces[0]) == 4:  # write tetrahedral mesh
                ntets = faces.shape[0]
                outF.write('  <mesh celltype="tetrahedron" dim="3">\n')
                outF.write('    <vertices size="' + str(ncoords) + '">\n')
                for i in range(ncoords):
                    x, y, z = vertices[i]
                    outF.write('      <vertex index="'+str(i)+'" x="'+x+'" y="'+y+'" z="'+z+'"/>\n')
                outF.write('    </vertices>\n')
                outF.write('    <cells size="' + str(ntets) + '">\n')
                for i in range(ntets):
                    v0, v1, v2, v3 = faces[i]
                    outF.write('     <tetrahedron index="'+str(i)
                               + '" v0="'+v0+'" v1="'+v1+'" v2="'+v2+'" v3="'+v3+'"/>\n')

            elif len(faces[0]) == 3:  # write triangle mesh
                ntri = faces.shape[0]
                outF.write('  <mesh celltype="triangle" dim="2">\n')
                outF.write('    <vertices size="' + str(ncoords) + '">\n')
                for i in range(ncoords):
                    x, y, _ = vertices[i]
                    outF.write('      <vertex index="'+str(i)+'" x="'+x+'" y="'+y+'"/>\n')
                outF.write('    </vertices>\n')
                outF.write('    <cells size="' + str(ntri) + '">\n')
                for i in range(ntri):
                    v0, v1, v2 = faces[i]
                    outF.write('     <triangle index="'+str(i)+'" v0="'+v0+'" v1="'+v1+'" v2="'+v2+'"/>\n')

            outF.write("    </cells>\n")
            outF.write("  </mesh>\n")
            outF.write("</dolfin>\n")
        return objct

    else:
        vedo.logger.error(f"Unknown format {fileoutput}, file not saved")
        return objct

    try:
        if binary:
            writer.SetFileTypeToBinary()
        else:
            writer.SetFileTypeToASCII()
    except AttributeError:
        pass

    try:
        writer.SetInputData(obj)
        writer.SetFileName(fileoutput)
        writer.Write()
    except:
        vedo.logger.error(f"could not save {fileoutput}")
    return objct

def save(obj: Any, fileoutput="out.png", binary=True) -> Any:
    """Save an object to file. Same as `write()`."""
    return write(obj, fileoutput, binary)

def read(obj: Any, unpack=True, force=False) -> Any:
    """Read an object from file. Same as `load()`."""
    return load(obj, unpack, force)

###############################################################################
def export_window(fileoutput: Union[str, os.PathLike], binary=False, plt=None) -> "vedo.Plotter":
    """
    Exporter which writes out the rendered scene into an HTML, X3D or Numpy file.

    Example:
        - [export_x3d.py](https://github.com/marcomusy/vedo/tree/master/examples/other/export_x3d.py)

        Check out the HTML generated webpage [here](https://vedo.embl.es/examples/embryo.html).

        <img src='https://user-images.githubusercontent.com/32848391/57160341-c6ffbd80-6de8-11e9-95ff-7215ce642bc5.jpg' width="600"/>

    .. note::
        the rendering window can also be exported to `numpy` file `scene.npz`
        by pressing `E` key at any moment during visualization.
    """
    fileoutput = str(fileoutput)
    if plt is None:
        plt = vedo.plotter_instance

    fr = fileoutput.lower()
    ####################################################################
    if fr.endswith(".npy") or fr.endswith(".npz"):
        _export_npy(plt, fileoutput)

    ####################################################################
    elif fr.endswith(".x3d"):
        # obj = plt.get_actors()
        # if plt.axes_instances:
        #     obj.append(plt.axes_instances[0])

        # for a in obj:
        #     if isinstance(a, Assembly):
        #         plt.remove(a)
        #         plt.add(a.unpack())

        plt.render()

        exporter = vtki.new("X3DExporter")
        exporter.SetBinary(binary)
        exporter.FastestOff()
        exporter.SetInput(plt.window)
        exporter.SetFileName(fileoutput)
        # exporter.WriteToOutputStringOn()
        exporter.Update()
        exporter.Write()

        wsize = plt.window.GetSize()
        x3d_html = _x3d_html_template.replace("~fileoutput", fileoutput)
        x3d_html = x3d_html.replace("~width",  str(wsize[0]))
        x3d_html = x3d_html.replace("~height", str(wsize[1]))
        with open(fileoutput.replace(".x3d", ".html"), "w", encoding="UTF-8") as outF:
            outF.write(x3d_html)

    ####################################################################
    elif fr.endswith(".html"):
        savebk = vedo.notebook_backend
        vedo.notebook_backend = "k3d"
        vedo.settings.default_backend = "k3d"
        # acts = plt.get_actors()
        plt = vedo.backends.get_notebook_backend(plt.objects)

        with open(fileoutput, "w", encoding="UTF-8") as fp:
            fp.write(plt.get_snapshot())

        vedo.notebook_backend = savebk
        vedo.settings.default_backend = savebk

    else:
        vedo.logger.error(f"export extension {fr.split('.')[-1]} is not supported")

    return plt

#########################################################################
def to_numpy(act: Any) -> dict:
    """Encode a vedo object to numpy format."""

    ########################################################
    def _fillcommon(obj, adict):
        adict["filename"] = obj.filename
        adict["name"] = obj.name
        adict["time"] = obj.time
        adict["rendered_at"] = obj.rendered_at
        try:
            adict["transform"] = obj.transform.matrix
        except AttributeError:
            adict["transform"] = np.eye(4)

    ####################################################################
    try:
        obj = act.retrieve_object()
    except AttributeError:
        obj = act

    adict = {}
    adict["type"] = "unknown"

    ######################################################## Points/Mesh
    if isinstance(obj, (Points, vedo.UnstructuredGrid)):
        adict["type"] = "Mesh"
        _fillcommon(obj, adict)

        if isinstance(obj, vedo.UnstructuredGrid):
            # adict["type"] = "UnstructuredGrid"
            # adict["cells"] = obj.cells_as_flat_array
            poly = obj._actor.GetMapper().GetInput()
            mapper = obj._actor.GetMapper()
        else:
            poly = obj.dataset
            mapper = obj.mapper

        adict["points"] = obj.vertices.astype(float)

        adict["cells"] = None
        if poly.GetNumberOfPolys():
            adict["cells"] = obj.cells_as_flat_array

        adict["lines"] = None
        if poly.GetNumberOfLines():
            adict["lines"] = obj.lines#_as_flat_array

        adict["pointdata"] = {}
        for iname in obj.pointdata.keys():
            if "normals" in iname.lower():
                continue
            adict["pointdata"][iname] = obj.pointdata[iname]

        adict["celldata"] = {}
        for iname in obj.celldata.keys():
            if "normals" in iname.lower():
                continue
            adict["celldata"][iname] = obj.celldata[iname]
        
        adict["metadata"] = {}
        for iname in obj.metadata.keys():
            adict["metadata"][iname] = obj.metadata[iname]

        # NEW in vedo 5.0
        adict["scalar_mode"] = mapper.GetScalarMode()
        adict["array_name_to_color_by"] = mapper.GetArrayName()
        adict["color_mode"] = mapper.GetColorMode()
        adict["interpolate_scalars_before_mapping"] = mapper.GetInterpolateScalarsBeforeMapping()
        adict["use_lookup_table_scalar_range"] = mapper.GetUseLookupTableScalarRange()
        adict["scalar_range"] = mapper.GetScalarRange()
        adict["scalar_visibility"] = mapper.GetScalarVisibility()
        adict["pickable"] = obj.actor.GetPickable()
        adict["dragable"] = obj.actor.GetDragable()

        # adict["color_map_colors"]  = mapper.GetColorMapColors()   #vtkUnsignedCharArray
        # adict["color_coordinates"] = mapper.GetColorCoordinates() #vtkFloatArray
        texmap = mapper.GetColorTextureMap()  #vtkImageData
        if texmap:
            adict["color_texture_map"] = vedo.Image(texmap).tonumpy()
            # print("color_texture_map", adict["color_texture_map"].shape)

        adict["texture_array"] = None
        texture = obj.actor.GetTexture()
        if texture:
            adict["texture_array"] = vedo.Image(texture.GetInput()).tonumpy()
            adict["texture_interpolate"] = texture.GetInterpolate()
            adict["texture_repeat"] = texture.GetRepeat()
            adict["texture_quality"] = texture.GetQuality()
            adict["texture_color_mode"] = texture.GetColorMode()
            adict["texture_mipmap"] = texture.GetMipmap()
            adict["texture_blending_mode"] = texture.GetBlendingMode()
            adict["texture_edge_clamp"] = texture.GetEdgeClamp()
            adict["texture_border_color"] = texture.GetBorderColor()
            # print("tonumpy: texture", obj.name, adict["texture_array"].shape)

        adict["LUT"] = None
        adict["LUT_range"] = None
        lut = mapper.GetLookupTable()
        if lut:
            nlut = lut.GetNumberOfTableValues()
            lutvals = []
            for i in range(nlut):
                v4 = lut.GetTableValue(i)  # (r, g, b, alpha)
                lutvals.append(v4)
            adict["LUT"] = np.array(lutvals, dtype=np.float32)
            adict["LUT_range"] = np.array(lut.GetRange())

        prp = obj.properties
        adict["alpha"] = prp.GetOpacity()
        adict["representation"] = prp.GetRepresentation()
        adict["pointsize"] = prp.GetPointSize()

        adict["linecolor"] = None
        adict["linewidth"] = None
        adict["edge_visibility"] = prp.GetEdgeVisibility() # new in vedo 5.0
        if prp.GetEdgeVisibility():
            adict["linewidth"] = prp.GetLineWidth()
            adict["linecolor"] = prp.GetEdgeColor()

        adict["ambient"] = prp.GetAmbient()
        adict["diffuse"] = prp.GetDiffuse()
        adict["specular"] = prp.GetSpecular()
        adict["specularpower"] = prp.GetSpecularPower()
        adict["specularcolor"] = prp.GetSpecularColor()
        adict["shading"] = prp.GetInterpolation()  # flat phong..:
        adict["color"] = prp.GetColor()
        adict["lighting_is_on"] = prp.GetLighting()
        adict["backcolor"] = None
        if obj.actor.GetBackfaceProperty():
            adict["backcolor"] = obj.actor.GetBackfaceProperty().GetColor()

    ######################################################## Volume
    elif isinstance(obj, Volume):
        adict["type"] = "Volume"
        _fillcommon(obj, adict)
        adict["array"] = obj.tonumpy()
        adict["mode"] = obj.mode()
        adict["spacing"] = obj.spacing()
        adict["origin"] = obj.origin()        

        prp = obj.properties
        ctf = prp.GetRGBTransferFunction()
        otf = prp.GetScalarOpacity()
        gotf = prp.GetGradientOpacity()
        smin, smax = ctf.GetRange()
        xs = np.linspace(smin, smax, num=256, endpoint=True)
        cols, als, algrs = [], [], []
        for x in xs:
            cols.append(ctf.GetColor(x))
            als.append(otf.GetValue(x))
            if gotf:
                algrs.append(gotf.GetValue(x))
        adict["color"] = cols
        adict["alpha"] = als
        adict["alphagrad"] = algrs

    ######################################################## Image
    elif isinstance(obj, Image):
        adict["type"] = "Image"
        _fillcommon(obj, adict)
        adict["array"] = obj.tonumpy()
        adict["scale"] = obj.actor.GetScale()
        adict["position"] = obj.actor.GetPosition()
        adict["orientation"] = obj.actor.GetOrientation()
        adict['origin'] = obj.actor.GetOrigin()
        adict["alpha"] = obj.alpha()

    ######################################################## Text2D
    elif isinstance(obj, vedo.Text2D):
        adict["type"] = "Text2D"
        adict["rendered_at"] = obj.rendered_at
        adict["text"] = obj.text()
        adict["position"] = obj.GetPosition()
        adict["color"] = obj.properties.GetColor()
        adict["font"] = obj.fontname
        adict["size"] = obj.properties.GetFontSize() / 22.5
        adict["bgcol"] = obj.properties.GetBackgroundColor()
        adict["alpha"] = obj.properties.GetBackgroundOpacity()
        adict["frame"] = obj.properties.GetFrame()
    
    ######################################################## Assembly
    elif isinstance(obj, Assembly):
        adict["type"] = "Assembly"
        _fillcommon(obj, adict)
        adict["parts"] = []
        for a in obj.unpack():
            adict["parts"].append(to_numpy(a))

    else:
        # vedo.logger.warning(f"to_numpy: cannot export object of type {type(obj)}")
        pass

    return adict


#########################################################################
def _export_npy(plt, fileoutput="scene.npz") -> None:
    
    fileoutput = str(fileoutput)

    sdict = {}
    sdict["shape"] = plt.shape
    sdict["sharecam"] = plt.sharecam
    sdict["camera"] = dict(
        pos=plt.camera.GetPosition(),
        focal_point=plt.camera.GetFocalPoint(),
        viewup=plt.camera.GetViewUp(),
        distance=plt.camera.GetDistance(),
        clipping_range=plt.camera.GetClippingRange(),
        parallel_scale=plt.camera.GetParallelScale(),
    )
    sdict["position"] = plt.pos
    sdict["size"] = plt.size
    sdict["axes"] = 0
    sdict["title"] = plt.title
    sdict["backgrcol"] = colors.get_color(plt.renderer.GetBackground())
    sdict["backgrcol2"] = None
    if plt.renderer.GetGradientBackground():
        sdict["backgrcol2"] = plt.renderer.GetBackground2()
    sdict["use_depth_peeling"] = plt.renderer.GetUseDepthPeeling()
    sdict["use_parallel_projection"] = plt.camera.GetParallelProjection()
    sdict["default_font"] = vedo.settings.default_font

    sdict["objects"] = []

    actors = plt.get_actors(include_non_pickables=True)
    # this ^ also retrieves Actors2D
    allobjs = []
    for i, a in enumerate(actors):

        if not a.GetVisibility():
            continue

        try:
            ob = a.retrieve_object()
            # print("get_actors",[ob], ob.name)
            if isinstance(ob, Assembly):
                asse_scale = ob.GetScale()
                asse_pos = ob.GetPosition()
                asse_ori = ob.GetOrientation()
                asse_org = ob.GetOrigin()
                for elem in ob.unpack():
                    elem.name = f"ASSEMBLY{i}_{ob.name}_{elem.name}"
                    # elem.info.update({"assembly": ob.name}) # TODO
                    # elem.info.update({"assembly_scale": asse_scale})
                    # elem.info.update({"assembly_position": asse_pos})
                    # elem.info.update({"assembly_orientation": asse_ori})
                    # elem.info.update({"assembly_origin": asse_org})
                    elem.metadata["assembly"] = ob.name
                    elem.metadata["assembly_scale"] = asse_scale
                    elem.metadata["assembly_position"] = asse_pos
                    elem.metadata["assembly_orientation"] = asse_ori
                    elem.metadata["assembly_origin"] = asse_org
                    allobjs.append(elem)
            else:
                allobjs.append(ob)

        except AttributeError:
            # print()
            # vedo.logger.warning(f"Cannot retrieve object of type {type(a)}")
            pass

    for a in allobjs:
        # print("to_numpy(): dumping", [a], a.name)
        # try:
        npobj = to_numpy(a)
        sdict["objects"].append(npobj)
        # except AttributeError:
        #     vedo.logger.warning(f"Cannot export object of type {type(a)}")

    if fileoutput.endswith(".npz"):
        np.savez_compressed(fileoutput, vedo_scenes=[sdict])
    else:
        np.save(fileoutput, [sdict])


########################################################################
def import_window(fileinput: Union[str, os.PathLike]) -> Union["vedo.Plotter", None]:
    """
    Import a whole scene from a Numpy NPZ file.

    Returns:
        `vedo.Plotter` instance
    """
    fileinput = str(fileinput)

    if fileinput.endswith(".npy") or fileinput.endswith(".npz"):
        return _import_npy(fileinput)
    
    # elif ".obj" in fileinput.lower():
    #     meshes = load_obj(fileinput, mtl_file, texture_path)
    #     plt = vedo.Plotter()
    #     plt.add(meshes)
    #     return plt

    # elif fileinput.endswith(".h5") or fileinput.endswith(".hdf5"):
    #     return _import_hdf5(fileinput) # in store/file_io_HDF5.py

    return None


def load_obj(fileinput: Union[str, os.PathLike], mtl_file=None, texture_path=None) -> List[Mesh]:
    """
    Import a set of meshes from a OBJ wavefront file.

    Arguments:
        mtl_file : (str)
            MTL file for OBJ wavefront files
        texture_path : (str)
            path of the texture files directory

    Returns:
        `list(Mesh)`
    """
    fileinput = str(fileinput)

    window = vtki.vtkRenderWindow()
    window.SetOffScreenRendering(1)
    renderer = vtki.vtkRenderer()
    window.AddRenderer(renderer)

    importer = vtki.new("OBJImporter")
    importer.SetFileName(fileinput)
    if mtl_file is None:
        mtl_file = fileinput.replace(".obj", ".mtl").replace(".OBJ", ".MTL")
    if os.path.isfile(mtl_file):
        importer.SetFileNameMTL(mtl_file)
    if texture_path is None:
        texture_path = fileinput.replace(".obj", ".txt").replace(".OBJ", ".TXT")
    # since the texture_path may be a directory which contains textures
    if os.path.exists(texture_path):
        importer.SetTexturePath(texture_path)
    importer.SetRenderWindow(window)
    importer.Update()

    actors = renderer.GetActors()
    actors.InitTraversal()
    objs = []
    for _ in range(actors.GetNumberOfItems()):
        vactor = actors.GetNextActor()
        msh = Mesh(vactor)
        msh.name = "OBJMesh"
        tx = vactor.GetTexture()
        if tx:
            msh.texture(tx)
        objs.append(msh)
    return objs


##########################################################
def screenshot(filename="screenshot.png", scale=1, asarray=False) -> Union["vedo.Plotter", np.ndarray, None]:
    """
    Save a screenshot of the current rendering window.

    Alternatively, press key `Shift-S` in the rendering window to save a screenshot.
    You can also use keyword `screenshot` in `show(..., screenshot="pic.png")`.

    Arguments:
        scale : (int)
            Set image magnification as an integer multiplicative factor.
            E.g. setting a magnification of 2 produces an image twice as large,
            but 10x slower to generate.
        asarray : (bool)
            Return a numpy array of the image
    """
    filename = str(filename)
    # print("calling screenshot", filename, scale, asarray)

    if not vedo.plotter_instance or not vedo.plotter_instance.window:
        # vedo.logger.error("in screenshot(), rendering window is not present, skip.")
        return vedo.plotter_instance  ##########
    
    if vedo.plotter_instance.renderer:
        vedo.plotter_instance.renderer.ResetCameraClippingRange()

    if asarray and scale == 1 and not vedo.plotter_instance.offscreen:
        nx, ny = vedo.plotter_instance.window.GetSize()
        arr = vtki.vtkUnsignedCharArray()
        vedo.plotter_instance.window.GetRGBACharPixelData(0, 0, nx-1, ny-1, 0, arr)
        narr = vedo.vtk2numpy(arr).T[:3].T.reshape([ny, nx, 3])
        narr = np.flip(narr, axis=0)
        return narr  ##########

    ###########################
    if filename.endswith(".pdf"):
        writer = vtki.new("GL2PSExporter")
        writer.SetRenderWindow(vedo.plotter_instance.window)
        writer.Write3DPropsAsRasterImageOff()
        writer.SilentOn()
        writer.SetSortToBSP()
        writer.SetFileFormatToPDF()
        writer.SetFilePrefix(filename.replace(".pdf", ""))
        writer.Write()
        return vedo.plotter_instance  ##########

    elif filename.endswith(".svg"):
        writer = vtki.new("GL2PSExporter")
        writer.SetRenderWindow(vedo.plotter_instance.window)
        writer.Write3DPropsAsRasterImageOff()
        writer.SilentOn()
        writer.SetSortToBSP()
        writer.SetFileFormatToSVG()
        writer.SetFilePrefix(filename.replace(".svg", ""))
        writer.Write()
        return vedo.plotter_instance  ##########

    elif filename.endswith(".eps"):
        writer = vtki.new("GL2PSExporter")
        writer.SetRenderWindow(vedo.plotter_instance.window)
        writer.Write3DPropsAsRasterImageOff()
        writer.SilentOn()
        writer.SetSortToBSP()
        writer.SetFileFormatToEPS()
        writer.SetFilePrefix(filename.replace(".eps", ""))
        writer.Write()
        return vedo.plotter_instance  ##########

    if settings.screeshot_large_image:
        w2if = vtki.new("RenderLargeImage")
        w2if.SetInput(vedo.plotter_instance.renderer)
        w2if.SetMagnification(scale)
    else:
        w2if = vtki.new("WindowToImageFilter")
        w2if.SetInput(vedo.plotter_instance.window)
        if hasattr(w2if, "SetScale"):
            w2if.SetScale(int(scale), int(scale))
        if settings.screenshot_transparent_background:
            w2if.SetInputBufferTypeToRGBA()
        w2if.ReadFrontBufferOff()  # read from the back buffer
    w2if.Update()

    if asarray:
        pd = w2if.GetOutput().GetPointData()
        npdata = utils.vtk2numpy(pd.GetArray("ImageScalars"))
        # npdata = npdata[:, [0, 1, 2]]  # remove alpha channel, issue #1199
        ydim, xdim, _ = w2if.GetOutput().GetDimensions()
        npdata = npdata.reshape([xdim, ydim, -1])
        npdata = np.flip(npdata, axis=0)
        return npdata ###########################

    if filename.lower().endswith(".png"):
        writer = vtki.new("PNGWriter")
        writer.SetFileName(filename)
        writer.SetInputData(w2if.GetOutput())
        writer.Write()
    elif filename.lower().endswith(".jpg") or filename.lower().endswith(".jpeg"):
        writer = vtki.new("JPEGWriter")
        writer.SetFileName(filename)
        writer.SetInputData(w2if.GetOutput())
        writer.Write()
    else:  # add .png
        writer = vtki.new("PNGWriter")
        writer.SetFileName(filename + ".png")
        writer.SetInputData(w2if.GetOutput())
        writer.Write()
    return vedo.plotter_instance


def ask(*question, **kwarg) -> str:
    """
    Ask a question from command line. Return the answer as a string.
    See function `colors.printc()` for the description of the keyword options.

    Arguments:
        options : (list)
            a python list of possible answers to choose from.
        default : (str)
            the default answer when just hitting return.

    Example:
    ```python
    import vedo
    res = vedo.ask("Continue?", options=['Y','n'], default='Y', c='g')
    print(res)
    ```
    """
    kwarg.update({"end": " "})
    if "invert" not in kwarg:
        kwarg.update({"invert": True})
    if "box" in kwarg:
        kwarg.update({"box": ""})

    options = kwarg.pop("options", [])
    default = kwarg.pop("default", "")
    if options:
        opt = "["
        for o in options:
            opt += o + "/"
        opt = opt[:-1] + "]"
        colors.printc(*question, opt, **kwarg)
    else:
        colors.printc(*question, **kwarg)

    try:
        resp = input()
    except Exception:
        resp = ""
        return resp

    if options:
        if resp not in options:
            if default and str(repr(resp)) == "''":
                return default
            colors.printc("Please choose one option in:", opt, italic=True, bold=False)
            kwarg["options"] = options
            return ask(*question, **kwarg)  # ask again
    return resp


##############################################################################################
class Video:
    """
    Generate a video from a rendering window.
    """

    def __init__(self, name="movie.mp4", duration=None, fps=24, scale=1, backend="imageio"):
        """
        Class to generate a video from the specified rendering window.
        Program `ffmpeg` is used to create video from each generated frame.

        Arguments:
            name : (Union[str, os.PathLike])
                name of the output file.
            duration : (float)
                set the total `duration` of the video and recalculates `fps` accordingly.
            fps : (int)
                set the number of frames per second.
            scale : (int)
                set the image magnification as an integer multiplicative factor.
            backend : (str)
                the backend engine to be used `['imageio', 'ffmpeg', 'cv']`

        Examples:
            - [make_video.py](https://github.com/marcomusy/vedo/tree/master/examples/other/make_video.py)

            ![](https://user-images.githubusercontent.com/32848391/50739007-2bfc2b80-11da-11e9-97e6-620a3541a6fa.jpg)
        """
        self.name = str(name)
        self.duration = duration
        self.backend = backend
        self.fps = float(fps)
        self.command = "ffmpeg -loglevel panic -y -r"
        self.options = "-b:v 8000k"
        self.scale = scale

        self.frames = []
        self.tmp_dir = TemporaryDirectory()
        self.get_filename = lambda x: os.path.join(self.tmp_dir.name, x)
        colors.printc(":video:  Video file", self.name, "is open... ", c="m", end="")

    def add_frame(self) -> "Video":
        """Add frame to current video."""
        fr = self.get_filename(str(len(self.frames)) + ".png")
        screenshot(fr, scale=self.scale)
        self.frames.append(fr)
        return self

    def pause(self, pause=0) -> "Video":
        """Insert a `pause`, in seconds."""
        fr = self.frames[-1]
        n = int(self.fps * pause)
        for _ in range(n):
            fr2 = self.get_filename(str(len(self.frames)) + ".png")
            self.frames.append(fr2)
            os.system("cp -f %s %s" % (fr, fr2))
        return self

    def action(self, elevation=(0, 80), azimuth=(0, 359), cameras=(), resetcam=False) -> "Video":
        """
        Automatic shooting of a static scene by specifying rotation and elevation ranges.

        Arguments:
            elevation : list
                initial and final elevation angles
            azimuth_range : list
                initial and final azimuth angles
            cameras : list
                list of cameras to go through, each camera can be dictionary or a vtkCamera
        """
        if not self.duration:
            self.duration = 5

        plt = vedo.plotter_instance
        if not plt:
            vedo.logger.error("No vedo plotter found, cannot make video.")
            return self
        n = int(self.fps * self.duration)

        cams = []
        for cm in cameras:
            cams.append(utils.camera_from_dict(cm))
        nc = len(cams)

        plt.show(resetcam=resetcam, interactive=False)

        if nc:
            for i in range(n):
                plt.move_camera(cams, i / (n-1))
                plt.render()
                self.add_frame()

        else:  ########################################

            for i in range(n):
                plt.camera.Elevation((elevation[1] - elevation[0]) / n)
                plt.camera.Azimuth((azimuth[1] - azimuth[0]) / n)
                plt.render()
                self.add_frame()

        return self

    def close(self) -> None:
        """
        Render the video and write it to file.
        """
        if self.duration:
            self.fps = int(len(self.frames) / float(self.duration) + 0.5)
            colors.printc("recalculated fps:", self.fps, c="m", end="")
        else:
            self.fps = int(self.fps)

        ########################################
        if self.backend == "ffmpeg":
            out = os.system(
                self.command
                + " "
                + str(self.fps)
                + " -i "
                + f"'{self.tmp_dir.name}'"
                + os.sep
                + "%01d.png "
                + self.options
                + " "
                + f"'{self.name}'"
            )
            if out:
                vedo.logger.error(f":noentry: backend {self.backend} returning error: {out}")
            else:
                colors.printc(f":save: saved to {self.name}", c="m")

        ########################################
        elif "cv" in self.backend:
            try:
                import cv2
            except ImportError:
                vedo.logger.error("opencv is not installed")
                return

            cap = cv2.VideoCapture(os.path.join(self.tmp_dir.name, "%1d.png"))
            fourcc = cv2.VideoWriter_fourcc(*"mp4v")
            if vedo.plotter_instance:
                w, h = vedo.plotter_instance.window.GetSize()
                writer = cv2.VideoWriter(self.name, fourcc, self.fps, (w, h), True)
            else:
                vedo.logger.error("No vedo plotter found, cannot make video.")
                return

            while True:
                ret, frame = cap.read()
                if not ret:
                    break
                writer.write(frame)

            cap.release()
            writer.release()

        ########################################
        elif "imageio" in self.backend:
            try:
                import imageio
            except ImportError:
                vedo.logger.error("Please install imageio with:\n pip install imageio[ffmpeg]")
                return

            if self.name.endswith(".mp4"):
                writer = imageio.get_writer(self.name, fps=self.fps)
            elif self.name.endswith(".gif"):
                writer = imageio.get_writer(self.name, mode="I", duration=1 / self.fps)
            elif self.name.endswith(".webm"):
                writer = imageio.get_writer(self.name, format="webm", fps=self.fps)
            else:
                vedo.logger.error(f"Unknown format of {self.name}.")
                return

            for f in utils.humansort(self.frames):
                image = imageio.v3.imread(f)
                try:
                    writer.append_data(image)
                except TypeError:
                    vedo.logger.error(f"Could not append data to video {self.name}")
                    vedo.logger.error("Please install imageio with: pip install imageio[ffmpeg]")
                    break
            try:
                writer.close()
                colors.printc(f"... saved as {self.name}", c="m")
            except:
                colors.printc(f":noentry: Could not save video {self.name}", c="r")

        # finalize cleanup
        self.tmp_dir.cleanup()

    def split_frames(self, output_dir="video_frames", prefix="frame_", format="png") -> None:
        """Split an existing video file into frames."""
        try:
            import imageio
        except ImportError:
            vedo.logger.error("\nPlease install imageio with:\n pip install imageio")
            return

        # Create the output directory if it doesn't exist
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        # Create a reader object to read the video
        reader = imageio.get_reader(self.name)

        # Loop through each frame of the video and save it as image
        print()
        for i, frame in utils.progressbar(
            enumerate(reader), title=f"writing {format} frames", c="m", width=20
        ):
            output_file = os.path.join(output_dir, f"{prefix}{str(i).zfill(5)}.{format}")
            imageio.imwrite(output_file, frame, format=format)



from vedo import *
lines = load("https://discourse.vtk.org/uploads/short-url/nC2RjJgTerpHKR0jD02Na6BRHVl.vtp")
for k in lines.pointdata.keys(): 
    print("array:", k, lines.pointdata[k].dtype)
lines.cmap("rainbow", "cluster_idx")
show(lines, axes=1, bg='blackboard')


"""Dijkstra algorithm to compute the graph geodesic.
Take as input a polygonal mesh and perform
a shortest path calculation between two vertices."""
from vedo import IcoSphere, Earth, show


msh = IcoSphere(r=1.02, subdivisions=4)
msh.wireframe().alpha(0.2)

path = msh.geodesic([0.349,-0.440,0.852], [-0.176,-0.962,0.302])
# path = msh.geodesic(36, 442) # use vertex indices

# printc(geo.pointdata["VertexIDs"])

show(Earth(), msh, path, __doc__, bg2='lb', viewup="z", zoom=1.3).close()



"""Make a static 2D clone copy of a mesh
and place it in the rendering window"""
from vedo import Mesh, dataurl, show

man3d = Mesh(dataurl+'man.vtk')
man3d.rotate_z(20).rotate_x(-70).scale(0.2)
man3d.c('darkgreen').lighting('glossy')

# Make a 2D snapshot of a 3D mesh
# The coordinate system options are
#     0. Displays
#     1. Normalized Display
#     2. Viewport (origin is the bottom-left corner of the window)
#     3. Normalized Viewport
#     4. View (origin is the center of the window)
#     5. World (anchor the 2d image to mesh)
# (returns a Actor2D)
man2d = man3d.clone2d().coordinate_system(4).pos([0.4,0.4])
man2d.c('red4')

show(man3d, man2d, __doc__, axes=1).close()




"""Streamlines originating from a set of seed points in space
subjected to a vectorial field defined on a small set of points."""
from vedo import *
import pandas as pd

data = "https://raw.githubusercontent.com/plotly/datasets/master/vortex.csv"
df = pd.read_csv(data)
pts  = np.c_[df['x'], df['y'], df['z']]
wind = np.c_[df['u'], df['v'], df['w']]

vpts = Points(pts, r=10)
vpts.pointdata["Wind"] = wind

# Convert points to a volume to create a domain for the streamlines
vol = vpts.tovolume(kernel='shepard', n=4, dims=(20,20,20))
vol_pts = vol.coordinates
iwind = vol.pointdata["Wind"] # interpolated wind
arrs = Arrows(vol_pts, vol_pts + iwind*0.5, alpha=0.2)

# Subsample the points to use as seed points
seeds = vpts.clone().subsample(0.2)

# Compute stream lines with Runge-Kutta integration
# vol.pointdata.select("Wind") # in case there are other vectors
streamlines = vol.compute_streamlines(seeds.points)
streamlines.pointdata["wind_intensity"] = mag(streamlines.pointdata["Wind"])
streamlines.cmap("Reds").add_scalarbar()
print(streamlines)
show(seeds, arrs, streamlines, __doc__, axes=1, viewup='z').close()

# Create a tube around the streamlines
streamtubes = Tubes(streamlines, r=0.01, vary_radius_by_scalar=True)
streamtubes.cmap("Reds").add_scalarbar()
print(streamtubes)
show(streamtubes, __doc__, axes=1, viewup='z').close()



from vedo import *

class SubMesh:
    """
    Cut out a submesh and glue it back, possibly with updated vertices, to the original mesh.
    The number and ordering of vertices in the resultant mesh are preserved.
    Class properties:
    * original_mesh
    * submesh: sub-mesh cut from original_mesh
    * mesh: the resultant mesh with submesh glued back
    * old_pids: indices of the submesh vertices which are also the mesh vertices
    * new_pids: indices of the new vertices added to submesh along the cut lines
    * cut: the pointcloud of the new vertices
    * dist2cut: distances from the original vertices in the submesh (old_pids) to the cut
    """
    def __init__(self, msh: Mesh, cut_fn_name: str, **kwargs):
        """
        :param msh: a Mesh
        :param cut_fn_name: Mesh method name to cut the Mesh
        :param kwargs: keyworded arguments to the cut meshod
        """
        self.original_mesh = msh
        self.mesh = msh.clone()
        self.mesh.pointdata['pids'] = np.arange(self.mesh.nvertices)
        self.submesh = getattr(self.mesh.clone(), cut_fn_name)(**kwargs)
        verts = Points(self.mesh.vertices)
        self.old_pids = []
        self.new_pids = []
        for i, v in enumerate(self.submesh.vertices):
            if Point(v).distance_to(verts) < 1e-3:
                self.old_pids.append(i)
            else:
                self.new_pids.append(i)
        self.cut = Points(self.submesh.vertices[self.new_pids])
        self.dist2cut = dict()

    def glue_(self, radius, align):
        """
        Glue submesh with possibly modified vertex positions back to the original mesh.
        :param radius: smoothing radius. The vertices of submesh which were originally
        at the distance smaller than radius, are interpolated between the original and new positions proportionally
        to the distance
        :param align: align the cut of submesh to the cut of the original mesh before gluing
        :return: mesh with submesh glued back
        """
        sm = self.submesh.clone()
        if align:
            sm.align_with_landmarks(self.submesh.vertices[self.new_pids], self.cut.vertices, rigid=True)
        if radius > 0:
            if len(self.dist2cut) == 0:  # pre-compute the distances for interactive gluing
                for i in self.old_pids:
                    pos = self.original_mesh.vertices[self.submesh.pointdata['pids'][i]]
                    self.dist2cut[i] = Point(pos).distance_to(self.cut).item()
            for i in self.old_pids:
                d = min(self.dist2cut[i] / radius, 1.)
                self.mesh.vertices[self.submesh.pointdata['pids'][i]] = (
                        d * sm.vertices[i] + (1-d) * self.original_mesh.vertices[self.submesh.pointdata['pids'][i]])
        else:
            for i in self.old_pids:
                self.mesh.vertices[self.submesh.pointdata['pids'][i]] = sm.vertices[i]

        self.mesh.pointdata.remove('pids')

    def glue(self, radius: float=0, mesh_col="wheat", align=False, interactive=False):
        """
        Glue submesh with possibly modified vertex positions back to the original mesh.
        :param radius: smoothing radius. The vertices of submesh which were originally
        at the distance smaller than radius, are interpolated between the original and new positions proportionally
        to the distance
        :param mesh_col: colour of the mesh in the plot
        :param align: align the cut of submesh to the cut of the original mesh before gluing
        :param interactive: open an interactive plot to adjust the smoothing radius
        :return: mesh with submesh glued back
        """
        self.glue_(radius=radius, align=align)
        if not interactive:
            return
        else:
            if len(self.dist2cut) == 0:  # pre-compute the distances for interactive gluing
                for i in self.old_pids:
                    pos = self.original_mesh.vertices[self.submesh.pointdata['pids'][i]]
                    self.dist2cut[i] = Point(pos).distance_to(self.cut).item()

            self.plt = Plotter()
            self.plt += self.mesh.c(mesh_col)

            def stitch(widget, event):
                self.glue_(radius=widget.value**2, align=align)
                self.plt -= self.mesh
                self.plt += self.mesh.c(mesh_col)

            self.plt.add_slider(
                stitch,
                value=radius,
                xmin=0,
                xmax=np.array(list(self.dist2cut.values())).max()**0.5 *2,
                pos="bottom",
                title="Smoothing radius",
            )
            self.plt.show(interactive=True).close()


S = Sphere(r=1, res=50).lw(1).flat()
box = Cube(side=1.5).wireframe()
cups = SubMesh(S, 'cut_with_box', bounds=box, invert=True)
cups.submesh.scale(1.2)  # alter the submesh
cups.glue(radius=0.2, mesh_col="coral", interactive=True)


man = Mesh(dataurl+"man.vtk").rotate_x(-90)
man.color('w').lw(1).flat()
cut_height = 1.20
head = SubMesh(man, 'cut_with_plane', origin=(0, cut_height, 0), normal=(0, 1, 0))
# modify the head:
head.submesh.scale(1.2, origin=(0,cut_height,0)).shift((0, 0.05, 0))
head.glue(interactive=True)


"""Scatter plot of 1M points with
assigned colors and transparencies.

Use mouse to zoom,
press r to reset,
press p to increase point size."""
from vedo import *

N = 1000000

x = np.random.rand(N)
y = np.random.rand(N)
RGBA = np.c_[x*255, y*255, np.zeros(N), y*255]

pts = np.array([x,y]).T
pts = Points(pts).point_size(1)
pts.pointcolors = RGBA

# use mouse to zoom, press r to reset
show(pts, __doc__, axes=1, mode="RubberBandZoom").close()



"""Simulate an elastic pendulum.
The trail is colored according to the velocity."""
import numpy as np
from scipy.integrate import solve_ivp
import matplotlib.pyplot as plt
from vedo import Plotter, Axes, Sphere, Spring, Image, mag, sin, cos
from vedo.addons import ProgressBarWidget


a = 2.0   # length of the pendulum
m = 0.5   # mass
k = 10.0  # constant of the spring
g = 9.81  # gravitational constant

# Define the system of ODEs
def system(t, z):
    x, dx_dt, y, dy_dt = z  # z = [x, x', y, y']
    dxdot_dt = (a+x) * dy_dt**2 - k/m * x + g * cos(y)
    dydot_dt = -2/(a+x) * dx_dt * dy_dt - g/(a+x) * sin(y)
    return [dx_dt, dxdot_dt, dy_dt, dydot_dt]


# Initial conditions: x(0), x'(0), y(0), y'(0)
initial_conditions = [0.0,   0.0,  0.4,   0.0]

# Time span for the solution
t_span = (0, 12)
t_eval = np.linspace(t_span[0], t_span[1], 500) # range to evaluate solution

# Solve the system numerically
solution = solve_ivp(system, t_span, initial_conditions, t_eval=t_eval)
t_values = solution.t
elong_values = solution.y[0]
theta_values = solution.y[2]

# Plot the results using matplotlib as a graph
fig = plt.figure()
plt.plot(t_values, elong_values, label="elongation(t)")
plt.plot(t_values, theta_values, label="angle(t)")
plt.xlabel("Time")
plt.legend()

# Animate the system using the solution of the ODE
plotter = Plotter(bg="blackboard", bg2="blue1", interactive=False)
pbw  = ProgressBarWidget(len(t_values))
axes = Axes(xrange=[-2, 2], yrange=[-a*2, 1], xygrid=0, xyframe_line=0, c="w")
img  = Image(fig).clone2d("top-right", size=0.5)
sphere = Sphere(r=0.3, c="red5").add_trail(c="k5", lw=4)
plotter.show(axes, sphere, img, pbw, __doc__)

for elong, theta in zip(elong_values, theta_values):
    x =  (a + elong) * sin(theta)
    y = -(a + elong) * cos(theta)
    spring = Spring([0, 0], [x, y])
    sphere.pos([x, y]).update_trail()

    # color the trail according to the lenght of each segment
    v = sphere.trail.vertices
    lenghts1 = np.array(v[1:])
    lenghts2 = np.array(v[:-1])
    lenghts = mag(lenghts1 - lenghts2) # lenght of each segment
    lenghts = np.append(lenghts, lenghts[-1])
    sphere.trail.cmap("Blues_r", lenghts, vmin=0, vmax=0.1)

    plotter.remove("Spring").add(spring).render()
    pbw.update()  # update progress bar

plotter.interactive()



"""The 3D-ruler axes and a goniometer"""
from vedo import *

settings.use_parallel_projection = True  # avoid parallax effects

mesh = Cone().rotate_y(90).pos([1, 2, 3])
mesh.c("steelblue")

# measure the angle formed by 3 points
gon = Goniometer(
    [-0.5, 1, 2], [2.5, 2, 2], [-0.5, 3, 3], 
    prefix=":alpha_c =~", lw=2, s=0.8
)

# show distance of any 2 points
rul = Ruler3D(
    (-0.5, 2, 1.9),
    (2.5, 2, 2.9),
    prefix="L_x =",
    units="μm",
    axis_rotation=90,
    tick_angle=70,
)

# make 3d rulers along the bounding box (similar to set axes=7)
ax3 = RulerAxes(mesh, units="μm")

show(mesh, gon, rul, ax3, __doc__, bg2="lb", viewup="z").close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
from vedo import printc, Text2D, Text3D, show, Plotter
from vedo import fonts, fonts_path, settings
import numpy as np

################################################################################## 2D
inred = Text2D(
    "°monospaced fonts are marked in red", c="r5", pos="bottom-center", font="VictorMono"
)
acts2d = [inred]

txt = "The quick fox jumps over the lazy dog. 1234567890 αβγδεθλμνπστφψω"
for i, f in enumerate(fonts):
    bg = None
    if f in ["Calco", "Glasgo", "SmartCouric", "VictorMono"]:
        bg = "red5"
    t = Text2D(f"{f}: {txt}", pos=(0.015, 1 - (i + 3) * 0.06), font=f, s=1.3, c="k", bg=bg)
    acts2d.append(t)

acts2d.append(Text2D("List of built-in fonts", pos="top-center", bg="k", s=1.3))
plt0a = show(acts2d, bg2="cornsilk", size=(1300, 800), interactive=False)

## online fonts:
acts2d = []
i = 0
for key, props in sorted(settings.font_parameters.items()):
    if props["islocal"]:
        continue
    if key in ("Justino2", "Justino3"):
        continue
    bg = None
    if props["mono"]:
        bg = "red5"
    t = Text2D(f"{key}: {txt}", pos=(0.015, 1 - (i + 2) * 0.03), font=key, c="k", bg=bg)
    acts2d.append(t)
    i += 1
plt0b = show(
    acts2d,
    Text2D("Additional fonts (https://vedo.embl.es/fonts)", pos="top-center", bg="k"),
    bg2="lb",
    size=(1300, 1350),
    pos=(1200, 0),
    new=True,
)

################################################################################ printout
for font in fonts:
    printc(font + " - available characters:", " " * 25, bold=1, invert=1)
    fontfile = os.path.join(fonts_path, font + ".npz")
    font_meshes = np.load(fontfile, allow_pickle=True)["font"][0]
    for k in font_meshes.keys():
        printc(k, end=" ")
    print()

printc("\n(use the above to copy&paste any char into your python script!)", italic=1)
printc("Symbols ~ ^ _ are reserved modifiers:", italic=1)
printc(" use ~ to add a short space, 1/4 of the default size,", italic=1)
printc(" use ^ and _ to start up/sub scripting, space terminates them.\n", italic=1)


################################################################################## 3D
# Symbols ~ ^ _ are reserved modifiers:
#  use ~ to add a short space, 1/4 of the default size,
#  use ^ and _ to start up/sub scripting, a space terminates them.
txt = """The quick fox jumps over the lazy dog.
Symbols: !@#$%&*()+=-{}[]:;|<>?/:euro1234567890
Units:  :delta=0.25E-03 ~μm, T_sea ~=~5.3~±0.7~:circC
LaTeX: :nabla:dotE=~4:pi~:rho, :nabla:timesE=~-1/c~~:partialB/:partialt
       ih~:partial/:partialt~:Psi = [-h^2 /2m:nabla^2  + V(r,t)]~:Psi(r,t)
       :DeltaE~=~h:nu, y = :Sigma_n ~A_n cos(:omega_n t+:delta_n ) sin(k_n x)
       :intx:dot~dx = :onehalf x:^2 + const.
       d^2 x^:mu  + :Gamma^:mu_:alpha:beta ~dx^:alpha ~dx^:beta  = 0
       -∇:^2u(x) = f(x) in Ω, u(x)~=~u_D (x) in :partial:Omega
"""

plt = Plotter(N=4, pos=(300, 0), size=(1600, 950))

cam = dict(
    pos=(3.99e5, 8.51e3, 6.47e5),
    focal_point=(2.46e5, 1.16e5, -9.24e3),
    viewup=(-0.0591, 0.983, 0.175),
    distance=6.82e5,
    clipping_range=(5.26e5, 8.92e5),
)

for i, fnt in enumerate(["Kanopus", "Normografo", "Theemim", "VictorMono"]):
    t = Text3D(txt, font=fnt, italic=0).c("darkblue").scale(12300)
    plt.at(i)
    plt.show(
        t,
        Text2D("Font: " + fnt, font=fnt, bg="r"),
        axes=dict(
            xtitle="my units for L_x  (:mum)",
            ytitle="my Y-axis with:na long description",
            title_font=fnt,
            label_font=fnt,
            digits=2,
        ),
        camera=cam,
        resetcam=not bool(i),
    )
plt.interactive().close()

plt0b.close()
plt0a.close()



"""Interactively slice a Volume along a plane.
Middle button + drag to slide the plane along the arrow"""
from vedo import *

normal = [0, 0, 1]
cmap = "gist_stern_r"

def func(w, _):
    c, n = pcutter.origin, pcutter.normal
    vslice = vol.slice_plane(c, n, autocrop=True).cmap(cmap)
    vslice.name = "Slice"
    plt.at(1).remove("Slice").add(vslice)


vol = Volume(dataurl + "embryo.slc").cmap(cmap)
vslice = vol.slice_plane(vol.center(), normal).cmap(cmap)
vslice.name = "Slice"

plt = Plotter(axes=0, N=2, bg="k", bg2="bb", interactive=False)
plt.at(0).show(vol, __doc__, zoom=1.5)

pcutter = PlaneCutter(
    vslice,
    normal=normal,
    alpha=0,
    c="white",
    padding=0,
    can_translate=False,
    can_scale=False,
)
pcutter.add_observer("interaction", func)
plt.at(1).add(pcutter)

plt.interactive()
plt.close()



from vedo import *


def func(widget, e):
    x = widget.value
    m = msh.clone()
    ids = m.find_cells_in_bounds(xbounds=(-10,x))
    m.delete_cells(ids)
    plt.remove("frog").add(m).render()


msh = Mesh("data/frog.obj").texture("data/frog.jpg")
msh.name = "frog"

plt = Plotter(axes=1)
plt.add_slider(func, xmin=-6, xmax=3, value=-6)
plt.show(msh)
plt.close()


from vedo import dataurl, Plotter, Volume


# Load a mesh and show it
vol = Volume(dataurl + "embryo.tif")
m0 = vol.isosurface(flying_edges=False).normalize()
m0.lw(1).c("violet")

# Smooth the mesh
m1 = m0.clone().smooth(niter=20)
m1.color("lg")

plt = Plotter(N=2)
plt.at(0).background("light blue")  # set first renderer color
plt.show(m0, "Original Mesh:")

plt.at(1)
plt.show("Mesh polygons are smoothed:", m1, viewup="z", zoom=1.5)
plt.interactive().close()



from vedo import UnstructuredGrid, Points, show, settings, utils

#####################################
def makeTetrahedron():
    """A tetrahedron"""
    pts = [
        (0, 0, 0),
        (1, 0, 0),
        (1, 1, 0),
        (0, 1, 1),
    ]
    cells = [[0, 1, 2, 3]]
    cellstypes = [10]
    ug = UnstructuredGrid([pts, cells, cellstypes])
    ug.c('w', 0.25).lw(2).lighting("off")
    return ug


#####################################
def makeHexahedron():
    """A regular hexagon (cube) with all faces square and three squares around
    each vertex is created below.
    Setup the coordinates of eight points (the two faces must be in
    counter clockwise order as viewed from the outside).
    """
    pts = [
        (0.0, 0.0, 0.0),
        (1.0, 0.0, 0.0),
        (1.0, 1.0, 0.0),
        (0.0, 1.0, 0.0),
        (0.0, 0.0, 1.0),
        (1.0, 0.0, 1.0),
        (1.0, 1.0, 1.0),
        (0.0, 1.0, 1.0),
    ]
    cells = [[0, 1, 2, 3, 4, 5, 6, 7]]
    cellstypes = [12]
    ug = UnstructuredGrid([pts, cells, cellstypes])
    ug.c('w', 0.25).lw(2).lighting("off")
    return ug


#####################################
def makePyramid():
    """Make a regular square pyramid"""
    pts = [
        [1.0, 1.0, 0.0],
        [-1.0, 1.0, 0.0],
        [-1.0, -1.0, 0.0],
        [1.0, -1.0, 0.0],
        [0.0, 0.0, 1.0],
    ]
    cells = [[0, 1, 2, 3, 4]]
    cellstypes = [14]
    ug = UnstructuredGrid([pts, cells, cellstypes])
    ug.c('w', 0.25).lw(2).lighting("off")
    return ug


#####################################
def makeWedge():
    """A wedge consists of two triangular ends and three rectangular faces"""
    pts = [
        (0, 1, 0),
        (0, 0, 0),
        (0, 0.25, 0.25),
        (1, 1, 0),
        (1, 0.0, 0.0),
        (1, 0.25, 0.25),
    ]
    cells = [[0, 1, 2, 3, 4, 5]]
    cellstypes = [13]
    ug = UnstructuredGrid([pts, cells, cellstypes])
    ug.c('w', 0.25).lw(2).lighting("off")
    return ug


#####################################
def makeHexagonalPrism():
    """Hexagonal prism: a wedge with an hexagonal base.
    Be careful, the base face ordering is different from wedge.
    """
    pts = [
        (0.0, 0.0, 1.0),
        (1.0, 0.0, 1.0),
        (1.5, 0.5, 1.0),
        (1.0, 1.0, 1.0),
        (0.0, 1.0, 1.0),
        (-0.5, 0.5, 1.0),
        ####
        (0.0, 0.0, 0.0),
        (1.0, 0.0, 0.0),
        (1.5, 0.5, 0.0),
        (1.0, 1.0, 0.0),
        (0.0, 1.0, 0.0),
        (-0.5, 0.5, 0.0),
    ]
    cells = [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]]
    cellstypes = [16]
    ug = UnstructuredGrid([pts, cells, cellstypes])
    ug.c('w', 0.25).lw(2).lighting("off")
    return ug


#####################################
def makePentagonalPrism():
    """A 3D pentagonal prism: a wedge with an pentagonal base."""
    pts = [
        (11, 10, 10),
        (13, 10, 10),
        (14, 12, 10),
        (12, 14, 10),
        (10, 12, 10),
        (11, 10, 14),
        (13, 10, 14),
        (14, 12, 14),
        (12, 14, 14),
        (10, 12, 14),
    ]
    cells = [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]]
    cellstypes = [15]
    ug = UnstructuredGrid([pts, cells, cellstypes])
    ug.c('w', 0.25).lw(2).lighting("off")
    return ug


def makeQuadraticTetra():
    import vtk

    print(utils.vtkVersionIsAtLeast(9))
    aTetra = vtk.vtkQuadraticTetra()
    points = vtk.vtkPoints()

    pcoords = aTetra.GetParametricCoords()
    rng = vtk.vtkMinimalStandardRandomSequence()
    points.SetNumberOfPoints(aTetra.GetNumberOfPoints())
    rng.SetSeed(5070)  # for testing
    for i in range(0, aTetra.GetNumberOfPoints()):
        perturbation = [0.0] * 3
        for j in range(0, 3):
            rng.Next()
            perturbation[j] = rng.GetRangeValue(-0.1, 0.1)
        aTetra.GetPointIds().SetId(i, i)
        points.SetPoint(i, [pcoords[3 * i], pcoords[3 * i + 1], pcoords[3 * i + 2]])

    ug = vtk.vtkUnstructuredGrid()
    ug.SetPoints(points)
    ug.InsertNextCell(aTetra.GetCellType(), aTetra.GetPointIds())
    return UnstructuredGrid(ug)


#####################################
tetr = makeTetrahedron()
hexa = makeHexahedron()
pirm = makePyramid()
wedg = makeWedge()
hexp = makeHexagonalPrism()
penp = makePentagonalPrism()
# qtetr= makeQuadraticTetra()

settings.immediate_rendering = False  # faster for multi-renderers

show(
    [
        ["Tetrahedron", tetr, Points(tetr.vertices, r=15, c="o")],
        ["Hexahedron", hexa, Points(hexa.vertices, r=15, c="o")],
        ["Pyramid", pirm, Points(pirm.vertices, r=15, c="o")],
        ["Wedge", wedg, Points(wedg.vertices, r=15, c="o")],
        ["HexagonalPrism", hexp, Points(hexp.vertices, r=15, c="o")],
        ["PentagonalPrism", penp, Points(penp.vertices, r=15, c="o")],
    ],
    N=6,
    bg="blue3",
    sharecam=False,
    axes=1,
).close()



"""Slice a Volume with multiple planes
Make low values of the scalar completely transparent"""
from vedo import *

vol = Volume(dataurl+'embryo.slc')
vol.cmap('bone').alpha([0,0,0.5])

slices = []
for i in range(4):
    sl = vol.slice_plane(origin=[150,150,i*50+50], normal=(0,-1,1))
    slices.append(sl)

amap = [0, 1, 1, 1, 1]  # hide low value points giving them alpha 0
mslices = merge(slices) # merge all slices into a single Mesh
mslices.cmap('hot_r', alpha=amap).lighting('off')

# Add a 3d scalarbar to the scene
mslices.add_scalarbar3d()
# make 2d copy of it and move it to the center-right
mslices.scalarbar = mslices.scalarbar.clone2d("center-right", 0.15)

show(vol, mslices, __doc__, axes=1).close()



import numpy as np
from vedo import Cone, Sphere, merge, Volume, dataurl, utils
import vtk

print('\n\n---------------------------------')
print('vtkVersion', vtk.vtkVersion().GetVTKVersion())
print('---------------------------------')


#####################################
cone = Cone(res=48)
sphere = Sphere(res=24)

carr = cone.cell_centers().coordinates[:, 2]
parr = cone.vertices[:, 0]

cone.pointdata["parr"] = parr
cone.celldata["carr"] = carr

carr = sphere.cell_centers().coordinates[:, 2]
parr = sphere.vertices[:, 0]

sphere.pointdata["parr"] = parr
sphere.celldata["carr"] = carr

sphere.pointdata["pvectors"] = np.sin(sphere.vertices)

sphere.compute_elevation()

cone.compute_normals()
sphere.compute_normals()


###################################### test clone()
c2 = cone.clone()
print('clone()', cone.npoints, c2.npoints)
assert cone.npoints == c2.npoints
print('clone()', cone.ncells, c2.ncells)
assert cone.ncells == c2.ncells


###################################### test merge()
m = merge(sphere, cone)
print('merge()', m.npoints, cone.npoints + sphere.npoints)
assert m.npoints == cone.npoints + sphere.npoints
print('merge()', m.ncells, cone.ncells + sphere.ncells)
assert m.ncells == cone.ncells + sphere.ncells


###################################### inputdata
print('dataset', [cone.dataset], "vtk.vtkPolyData")
assert isinstance(cone.dataset, vtk.vtkPolyData)


###################################### mapper
print('mapper', [cone.mapper], "vtk.vtkPolyDataMapper")
assert isinstance(cone.mapper, vtk.vtkPolyDataMapper)


###################################### pickable
cone.pickable(False)
cone.pickable(True)
print('pickable', cone.pickable(), True)
assert cone.pickable()


###################################### pos
cone.pos(1,2,3)
print('pos', [1,2,3], cone.pos())
assert np.allclose([1,2,3], cone.pos())
cone.pos(5,6)
print('pos',[5,6,0], cone.pos())
assert np.allclose([5,6,0], cone.pos())


###################################### shift
cone.pos(5,6,7).shift(3,0,0)
print('shift',[8,6,7], cone.pos())
assert np.allclose([8,6,7], cone.pos(), atol=0.001)


###################################### x y z
cone.pos(10,11,12)
cone.x(1.1)
print('x y z',[1.1,11,12], cone.pos())
assert np.allclose([1.1,11,12], cone.pos(), atol=0.001)
cone.y(1.2)
print('x y z',[1.1,1.2,12], cone.pos())
assert np.allclose([1.1,1.2,12], cone.pos(), atol=0.001)
cone.z(1.3)
print('x y z',[1.1,1.2,1.3], cone.pos())
assert np.allclose([1.1,1.2,1.3], cone.pos(), atol=0.001)


###################################### rotate
cr = cone.pos(0,0,0).clone().rotate(90, axis=(0, 1, 0))
print('rotate', np.max(cr.vertices[:,2]) ,'<', 1.01)
assert np.max(cr.vertices[:,2]) < 1.01


###################################### orientation
cr = cone.pos(0,0,0).clone().reorient([0,0,1], (1, 1, 0))
print('orientation',np.max(cr.vertices[:,2]) ,'<', 1.01)
assert np.max(cr.vertices[:,2]) < 1.01

####################################### scale
cr.scale(5)
print('scale',np.max(cr.vertices[:,2]) ,'>', 4.98)
assert np.max(cr.vertices[:,2]) > 4.98


###################################### box
bx = cone.box()
print('box',bx.npoints, 24)
assert bx.npoints == 24
print('box',bx.clean().npoints , 8)
assert bx.clean().npoints == 8

###################################### transform
ct = cone.clone().rotate_x(10).rotate_y(10).rotate_z(10)
print('transform', [ct.transform.T], [vtk.vtkTransform])
assert isinstance(ct.transform.T, vtk.vtkTransform)

print('ntransforms',ct.transform.T.GetNumberOfConcatenatedTransforms())
assert ct.transform.T.GetNumberOfConcatenatedTransforms()


###################################### pointdata and celldata
arrnames = cone.pointdata.keys()
print('pointdata', arrnames, 'parr')
assert 'parr' in arrnames
arrnames = cone.celldata.keys()
print('celldata.keys', arrnames, 'carr')
assert 'carr' in arrnames


###################################### Get Point Data
arr = sphere.pointdata['parr']
print('pointdata',len(arr))
assert len(arr)
print('pointdata',np.max(arr) ,'>', .99)
assert np.max(arr) > .99

arr = sphere.celldata['carr']
print('celldata',[arr])
assert len(arr)
print('celldata',np.max(arr) ,'>', .99)
assert np.max(arr) > .99


######################################__add__
print('__add__', [cone+sphere], [vtk.vtkAssembly])
assert isinstance(cone+sphere, vtk.vtkAssembly)


###################################### vertices
s2 = sphere.clone()
pts = sphere.vertices
pts2 = pts + [1,2,3]
s2.vertices = pts2
pts3 = s2.vertices
print('vertices',sum(pts3-pts2))
assert np.allclose(pts2, pts3, atol=0.001)


###################################### cells
print('cells')
assert np.array(sphere.cells).shape == (2112, 3)


###################################### texture
st = sphere.clone().texture(dataurl+'textures/wood2.jpg')
print('texture test')
assert isinstance(st.actor.GetTexture(), vtk.vtkTexture)


###################################### delete_cells_by_point_index
sd = sphere.clone().delete_cells_by_point_index(range(100))
print('delete_cells_by_point_index',sd.npoints , sphere.npoints)
assert sd.npoints == sphere.npoints
print('delete_cells_by_point_index',sd.ncells ,'<', sphere.ncells)
assert sd.ncells < sphere.ncells


###################################### reverse
# this fails on some archs (see issue #185)
# lets comment it out temporarily
# sr = sphere.clone().reverse().cut_with_plane()
# print('DISABLED: reverse test', sr.npoints, 576)
# rev = vtk.vtkReverseSense()
# rev.SetInputData(sr.polydata())
# rev.Update()
# print('DISABLED: reverse vtk nr.pts, nr.cells')
# print(rev.GetOutput().GetNumberOfPoints(),sr.polydata().GetNumberOfPoints(),
#       rev.GetOutput().GetNumberOfCells(), sr.polydata().GetNumberOfCells())
# assert sr.npoints == 576


###################################### quantize
sq = sphere.clone().quantize(0.1)
print('quantize',sq.npoints , 834)
assert sq.npoints == 834


###################################### bounds
ss = sphere.clone().scale([1,2,3])
print('bounds',ss.xbounds())
assert np.allclose(ss.xbounds(), [-1,1], atol=0.01)
print('bounds',ss.ybounds())
assert np.allclose(ss.ybounds(), [-2,2], atol=0.01)
print('bounds',ss.zbounds())
assert np.allclose(ss.zbounds(), [-3,3], atol=0.01)


###################################### average_size
print('average_size', Sphere().scale(10).pos(1,3,7).average_size())
assert 9.9 < Sphere().scale(10).pos(1,3,7).average_size() < 10.1

print('diagonal_size',sphere.diagonal_size())
assert 3.3 < sphere.diagonal_size() < 3.5

print('center_of_mass',sphere.center_of_mass())
assert np.allclose(sphere.center_of_mass(), [0,0,0], atol=0.001)

print('volume',sphere.volume())
assert 4.1 < sphere.volume() < 4.2

print('area',sphere.area())
assert 12.5 < sphere.area() < 12.6


###################################### closest_point
pt = [12,34,52]
print('closest_point',sphere.closest_point(pt), [0.19883616, 0.48003298, 0.85441941])
assert np.allclose(sphere.closest_point(pt),
                   [0.19883616, 0.48003298, 0.85441941], atol=0.001)


###################################### find_cells_in_bounds
ics = sphere.find_cells_in_bounds(xbounds=(-0.5, 0.5))
print('find_cells_in',len(ics) , 1404)
assert len(ics) == 1576


######################################transformMesh
T = cone.clone().pos(35,67,87).transform
s3 = Sphere().apply_transform(T)
print('transformMesh',s3.center_of_mass(), (35,67,87))
assert np.allclose(s3.center_of_mass(), (35,67,87), atol=0.001)


######################################normalize
s3 = sphere.clone().pos(10,20,30).scale([7,8,9]).normalize()
print('normalize',s3.center_of_mass(), (10,20,30))
assert np.allclose(s3.center_of_mass(), (10,20,30), atol=0.001)
print('normalize',s3.average_size())
assert 0.9 < s3.average_size() < 1.1


###################################### crop
c2 = cone.clone().crop(left=0.5)
print('crop',np.min(c2.vertices[:,0]), '>', -0.001)
assert np.min(c2.vertices[:,0]) > -0.001


###################################### subdivide
s2 = sphere.clone().subdivide(4)
print('subdivide',s2.npoints , 270338)
assert s2.npoints == 270338


###################################### decimate
s2 = sphere.clone().decimate(0.2)
print('decimate',s2.npoints , 213)
assert s2.npoints == 213

###################################### vertex_normals
print('vertex_normals',sphere.vertex_normals[12], [9.97668684e-01, 1.01513637e-04, 6.82437494e-02])
assert np.allclose(sphere.vertex_normals[12], [9.97668684e-01, 1.01513637e-04, 6.82437494e-02], atol=0.001)


###################################### contains
print('is_inside',sphere.contains([0.1,0.2,0.3]))
assert Sphere().contains([0.1,0.2,0.3])

###################################### intersectWithLine (fails vtk7..)
# pts = sphere.intersectWithLine([-2,-2,-2], [2,3,4])
# print('intersectWithLine',pts[0])
# assert np.allclose(pts[0], [-0.8179885149002075, -0.522485613822937, -0.2269827425479889], atol=0.001)
# print('intersectWithLine',pts[1])
# assert np.allclose(pts[1], [-0.06572723388671875, 0.41784095764160156, 0.9014091491699219], atol=0.001)


############################################################################
############################################################################ Assembly
asse = cone+sphere

######################################
print('unpack',len(asse.unpack()) , 2)
assert len(asse.unpack()) ==2
print('unpack', asse.unpack(0).name)
assert asse.unpack(0) == cone
print('unpack',asse.unpack(1).name)
assert asse.unpack(1) == sphere
print('unpack',asse.diagonal_size(), 4.15)
assert 4.1 < asse.diagonal_size() < 4.2


############################################################################ Volume
X, Y, Z = np.mgrid[:30, :30, :30]
scalar_field = ((X-15)**2 + (Y-15)**2 + (Z-15)**2)/225
print('Test Volume, scalar min, max =', np.min(scalar_field), np.max(scalar_field))

vol = Volume(scalar_field)
volarr = vol.pointdata[0]

print('Volume',volarr.shape[0] , 27000)
assert volarr.shape[0] == 27000
print('Volume',np.max(volarr) , 3)
assert np.max(volarr) == 3
print('Volume',np.min(volarr) , 0)
assert np.min(volarr) == 0

###################################### isosurface
iso = vol.isosurface(1.0)
print('isosurface', iso.area())
assert 2540 < iso.area() <  3000


###################################### utils change of coords
from vedo import transformations
q = [5,2,3]
q = transformations.cart2spher(*q)
q = transformations.spher2cart(*q)
print("cart2spher spher2cart", q)
assert np.allclose(q, [5,2,3], atol=0.001)
q = transformations.cart2cyl(*q)
q = transformations.cyl2cart(*q)
print("cart2cyl cyl2cart", q)
assert np.allclose(q, [5,2,3], atol=0.001)
q = transformations.cart2cyl(*q)
q = transformations.cyl2spher(*q)
q = transformations.spher2cart(*q)
print("cart2cyl cyl2spher spher2cart", q)
assert np.allclose(q, [5,2,3], atol=0.001)
q = transformations.cart2spher(*q)
q = transformations.spher2cyl(*q)
q = transformations.cyl2cart(*q)
print("cart2spher spher2cyl cyl2cart", q)
assert np.allclose(q, [5,2,3], atol=0.001)

######################################
print("OK with test_actors")




from vedo import *

aline = Line(Circle().coordinates)
spline = Spline([(0,0,0), (1,1,1), (2,3,3), (1,1,4), (0,1,5)])
spline.lw(5)
pts = spline.coordinates

surfs = []
for i in range(1, len(pts)):
    p0, p1 = pts[i-1:i+1]
    surf = aline.sweep(p1 - p0)
    surfs.append(surf)
surface = merge(surfs, flag=True)
surface.c("gold").lw(0.1).pickable(True)

show(spline, surface, aline, axes=1).close()


"""Scale a mesh asymmetrically in one coordinate"""
from dolfin import *
from mshr import *

domain = Rectangle(Point(0.0, 0.0), Point(5.0, 0.01))
mesh = generate_mesh(domain, 20)
V = FunctionSpace(mesh, "CG", 2)

e = Expression("sin(2*pi*(x[0]*x[0]+x[1]*x[1]))", degree=2)
f = interpolate(e, V)

####################################################
from vedo.dolfin import plot

plt = plot(
    f,
    xtitle="y-coord is scaled by factor 100",
    scale_mesh_factors=(0.01, 1, 1),
    style=1,
    lw=0,
    warp_zfactor=0.001,
    scalarbar="horizontal",
    axes={"xtitle_offset": 0.2},
    text=__doc__,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os

__docformat__ = "google"


class Settings:
    """
    General settings to modify the global behavior and style.

    Example:
        ```python
        from vedo import settings, Cube
        settings.use_parallel_projection = True
        # settings["use_parallel_projection"] = True  # this is equivalent!
        Cube().color('g').show().close()
        ```

    List of available properties:

    ```python
    # Set the default font to be used for axes, comments etc.
    # Check out the available fonts at http://vedo.embl.es/fonts
    # For example:
    default_font = 'Normografo'
    # To customize the font parameters use:
    settings.font_parameters["Normografo"] = dict(
        mono=False,
        fscale=0.75,
        hspacing=1,
        lspacing=0.2,
        dotsep="~×",
        islocal=True,
    )
    # Where
    # mono    : if True all letters occupy the same space slot horizontally
    # fscale  : sets the general scaling factor for the size of the font
    # hspacing: horizontal stretching factor (affects both letters and words)
    # lspacing: horizontal spacing inbetween letters (not words)
    # dotsep  : a string of characters to be interpreted as dot separator
    # islocal : if locally stored in /fonts, otherwise it's on vedo.embl.es/fonts
    #
    # To run a demo try:
    # vedo --run fonts

    # Use this local folder to store downloaded files (default is ~/.cache/vedo)
    cache_directory = ".cache"

    # Palette number when using an integer to choose a color
    palette = 0

    # Options for saving window screenshots:
    screenshot_transparent_background = False
    screeshot_large_image = False # sometimes setting this to True gives better results

    # Enable tracking pipeline functionality:
    # allows to show a graph with the pipeline of action which let to a final object
    # this is achieved by calling "myobj.pipeline.show()" (a new window will pop up)
    self.enable_pipeline = True

    # Remember the last format used when creating new figures in vedo.pyplot
    # this is useful when creating multiple figures of the same kind
    # and avoid to specify the format each time in plot(..., like=...)
    remember_last_figure_format = False

    # Set up default mouse and keyboard callbacks
    enable_default_mouse_callbacks = True
    enable_default_keyboard_callbacks = True

    # Force single precsion of points coordinates.
    # Useful for very large point clouds and meshes. Default is True.
    force_single_precision_points = True

    # Progress bar delay before showing up [sec]
    progressbar_delay = 0.5

    # If False, when multiple renderers are present, render only once at the end
    immediate_rendering = True

    # In multirendering mode, show a grey frame margin (set width=0 to disable)
    renderer_frame_color = None
    renderer_frame_alpha = 0.5
    renderer_frame_width = 0.5
    renderer_frame_padding = 0.0001

    # In multirendering mode, set the position of the horizontal of vertical splitting [0,1]
    window_splitting_position = None

    # Gradient orientation mode for background window color
    # 0 = Vertical
    # 1 = Horizontal
    # 2 = Radial viewport farthest side
    # 3 = Radial viewport farthest corner
    background_gradient_orientation = 0

    # Enable / disable color printing by printc()
    enable_print_color = True

    # Smoothing options for points, lines and polygons
    point_smoothing = False
    line_smoothing = False
    polygon_smoothing = False

    # Turn on/off the automatic repositioning of lights as the camera moves
    light_follows_camera = False
    two_sided_lighting = True

    # Turn on/off rendering of translucent material with depth peeling technique
    use_depth_peeling = False
    alpha_bit_planes  = True   # options only active if useDepthPeeling=True
    multi_samples     = 8      # antialiasing multisample buffer
    max_number_of_peels= 4     # maximum number of rendering passes
    occlusion_ratio   = 0.0    # occlusion ratio, 0 = exact image.

    # Turn on/off nvidia FXAA post-process anti-aliasing, if supported
    use_fxaa = False           # either True or False

    # By default, the depth buffer is reset for each renderer
    #  If True, use the existing depth buffer
    preserve_depth_buffer = False

    # Use a polygon/edges offset to possibly resolve conflicts in rendering
    use_polygon_offset    = False
    polygon_offset_factor = 0.1
    polygon_offset_units  = 0.1

    # Interpolate scalars to render them smoothly
    interpolate_scalars_before_mapping = True

    # Set parallel projection On or Off (place camera to infinity, no perspective effects)
    use_parallel_projection = False

    # Set orientation type when reading TIFF files:
    # TOPLEFT  1 (row 0 top,    col 0 lhs)    TOPRIGHT 2 (row 0 top,    col 0 rhs)
    # BOTRIGHT 3 (row 0 bottom, col 0 rhs)    BOTLEFT  4 (row 0 bottom, col 0 lhs)
    # LEFTTOP  5 (row 0 lhs,    col 0 top)    RIGHTTOP 6 (row 0 rhs,    col 0 top)
    # RIGHTBOT 7 (row 0 rhs,    col 0 bottom) LEFTBOT  8 (row 0 lhs,    col 0 bottom)
    tiff_orientation_type = 1

    # Annotated cube axis type nr. 5 options:
    annotated_cube_color      = (0.75, 0.75, 0.75)
    annotated_cube_text_color = None # use default, otherwise specify a single color
    annotated_cube_text_scale = 0.2
    annotated_cube_texts      = ["right","left ", "front","back ", " top ", "bttom"]

    # Set the default backend for plotting in jupyter notebooks.
    # If a jupyter environment is detected, the default is automatically switched to "2d"
    default_backend = "vtk"

    # Automatically close the Plotter instance after show() in jupyter sessions
    # setting it to False will keep the current Plotter instance active
    backend_autoclose = True

    # Settings specific to the K3D backend in jupyter notebooks
    k3d_menu_visibility = True
    k3d_plot_height   = 512
    k3d_antialias     = True
    k3d_lighting      = 1.5
    k3d_camera_autofit= True
    k3d_grid_autofit  = True
    k3d_axes_color    = "gray4"
    k3d_axes_helper   = 1.0     # size of the small triad of axes on the bottom right
    k3d_point_shader  = "mesh"  # others are '3d', '3dSpecular', 'dot', 'flat'
    k3d_line_shader   = "thick" # others are 'flat', 'mesh'
    ```
    """

    # Restrict the attributes so accidental typos will generate an AttributeError exception
    __slots__ = [
        "default_font",
        "default_backend",
        "cache_directory",
        "palette",
        "remember_last_figure_format",
        "screenshot_transparent_background",
        "screeshot_large_image",
        "enable_default_mouse_callbacks",
        "enable_default_keyboard_callbacks",
        "enable_pipeline",
        "progressbar_delay",
        "immediate_rendering",
        "renderer_frame_color",
        "renderer_frame_alpha",
        "renderer_frame_width",
        "renderer_frame_padding",
        "force_single_precision_points",
        "point_smoothing",
        "line_smoothing",
        "polygon_smoothing",
        "light_follows_camera",
        "two_sided_lighting",
        "use_depth_peeling",
        "multi_samples",
        "alpha_bit_planes",
        "max_number_of_peels",
        "occlusion_ratio",
        "use_fxaa",
        "preserve_depth_buffer",
        "use_polygon_offset",
        "polygon_offset_factor",
        "polygon_offset_units",
        "interpolate_scalars_before_mapping",
        "use_parallel_projection",
        "background_gradient_orientation",
        "window_splitting_position",
        "tiff_orientation_type",
        "annotated_cube_color",
        "annotated_cube_text_color",
        "annotated_cube_text_scale",
        "annotated_cube_texts",
        "enable_print_color",
        "backend_autoclose",
        "k3d_menu_visibility",
        "k3d_plot_height",
        "k3d_antialias",
        "k3d_lighting",
        "k3d_camera_autofit",
        "k3d_grid_autofit",
        "k3d_axes_color",
        "k3d_axes_helper",
        "k3d_point_shader",
        "k3d_line_shader",
        "font_parameters",
    ]

    ############################################################
    # Dry run mode (for test purposes only)
    # 0 = normal
    # 1 = do not hold execution
    # 2 = do not hold execution and do not show any window
    dry_run_mode = 0

    ############################################################
    def __init__(self) -> None:

        self.default_backend = "vtk"
        try:
            # adapted from: https://stackoverflow.com/a/39662359/2912349
            shell = get_ipython().__class__.__name__
            if shell == 'ZMQInteractiveShell':
                self.default_backend = "2d"
        except NameError:
            pass

        self.default_font = "Normografo"

        self.enable_pipeline = True
        self.progressbar_delay = 0.5
        self.palette = 0
        self.remember_last_figure_format = False

        self.force_single_precision_points = True

        self.cache_directory = ".cache"  # "/vedo" is added automatically

        self.screenshot_transparent_background = False
        self.screeshot_large_image = False

        self.enable_default_mouse_callbacks = True
        self.enable_default_keyboard_callbacks = True
        self.immediate_rendering = True

        self.renderer_frame_color = None
        self.renderer_frame_alpha = 0.5
        self.renderer_frame_width = 0.5
        self.renderer_frame_padding = 0.0001
        self.background_gradient_orientation = 0

        self.point_smoothing = False
        self.line_smoothing = False
        self.polygon_smoothing = False

        self.light_follows_camera = False
        self.two_sided_lighting = True

        self.use_depth_peeling = False
        self.multi_samples = 8
        self.alpha_bit_planes = 1
        self.max_number_of_peels = 4
        self.occlusion_ratio = 0.1

        self.use_fxaa = False

        self.preserve_depth_buffer = False

        self.use_polygon_offset = True
        self.polygon_offset_factor = 0.1
        self.polygon_offset_units = 0.1

        self.interpolate_scalars_before_mapping = True

        self.use_parallel_projection = False

        self.window_splitting_position = None

        self.tiff_orientation_type = 1

        self.annotated_cube_color = (0.75, 0.75, 0.75)
        self.annotated_cube_text_color = None
        self.annotated_cube_text_scale = 0.2
        self.annotated_cube_texts = ["right", "left ", "front", "back ", " top ", "bttom"]

        self.enable_print_color = True

        self.backend_autoclose = True

        self.k3d_menu_visibility = True
        self.k3d_plot_height = 512
        self.k3d_antialias   = True
        self.k3d_lighting    = 1.5
        self.k3d_camera_autofit = True
        self.k3d_grid_autofit= True
        self.k3d_axes_color  = "k4"
        self.k3d_axes_helper = 1.0
        self.k3d_point_shader= "mesh"
        self.k3d_line_shader = "thick"

        self.font_parameters = dict(

            Normografo=dict(
                mono=False,
                fscale=0.75,
                hspacing=1,
                lspacing=0.2,
                dotsep="~×",
                islocal=True,
            ),
            Bongas=dict(
                mono=False,
                fscale=0.875,
                hspacing=0.52,
                lspacing=0.25,
                dotsep="·",
                islocal=True,
            ),
            Calco=dict(
                mono=True,
                fscale=0.8,
                hspacing=1,
                lspacing=0.1,
                dotsep="×",
                islocal=True,
            ),
            Comae=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=1,
                dotsep="~×",
                islocal=True,
            ),
            ComicMono=dict(
                mono=True,
                fscale=0.8,
                hspacing=1,
                lspacing=0.1,
                dotsep="x",
                islocal=False,
            ),
            Edo=dict(
                mono=False,
                fscale=0.75,
                hspacing=1,
                lspacing=0.2,
                dotsep="~x ",
                islocal=False,
            ),
            FiraMonoMedium=dict(
                mono=True,
                fscale=0.8,
                hspacing=1,
                lspacing=0.1,
                dotsep="×",
                islocal=False,
            ),
            FiraMonoBold=dict(
                mono=True,
                fscale=0.8,
                hspacing=1,
                lspacing=0.1,
                dotsep="×",
                islocal=False,
            ),
            Glasgo=dict(
                mono=True,
                fscale=0.75,
                lspacing=0.1,
                hspacing=1,
                dotsep="~×",
                islocal=True,
            ),
            Kanopus=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.15,
                hspacing=0.75,
                dotsep="~×",
                islocal=True,
            ),
            LogoType=dict(
                mono=False,
                fscale=0.75,
                hspacing=1,
                lspacing=0.2,
                dotsep="~×",
                islocal=False,
            ),
            Quikhand=dict(
                mono=False,
                fscale=0.8,
                hspacing=0.6,
                lspacing=0.15,
                dotsep="~~×~",
                islocal=True,
            ),
            SmartCouric=dict(
                mono=True,
                fscale=0.8,
                hspacing=1.05,
                lspacing=0.1,
                dotsep="×",
                islocal=True,
            ),
            Spears=dict(
                mono=False,
                fscale=0.8,
                hspacing=0.5,
                lspacing=0.2,
                dotsep="~×",
                islocal=False,
            ),
            Theemim=dict(
                mono=False,
                fscale=0.825,
                hspacing=0.52,
                lspacing=0.3,
                dotsep="~~×",
                islocal=True,
            ),
            VictorMono=dict(
                mono=True,
                fscale=0.725,
                hspacing=1,
                lspacing=0.1,
                dotsep="×",
                islocal=True,
            ),
            Justino1=dict(
                mono=True,
                fscale=0.725,
                hspacing=1,
                lspacing=0.1,
                dotsep="×",
                islocal=False,
            ),
            Justino2=dict(
                mono=True,
                fscale=0.725,
                hspacing=1,
                lspacing=0.1,
                dotsep="×",
                islocal=False,
            ),
            Justino3=dict(
                mono=True,
                fscale=0.725,
                hspacing=1,
                lspacing=0.1,
                dotsep="×",
                islocal=False,
            ),
            Calibri=dict(
                mono=False,
                fscale=0.75,
                hspacing=1,
                lspacing=0.2,
                dotsep="~×",
                islocal=False,
            ),
            Capsmall=dict(
                mono=False,
                fscale=0.8,
                hspacing=0.75,
                lspacing=0.15,
                dotsep="~×",
                islocal=False,
            ),
            Cartoons123=dict(
                mono=False,
                fscale=0.8,
                hspacing=0.75,
                lspacing=0.15,
                dotsep="x",
                islocal=False,
            ),
            Darwin=dict(
                mono=False,
                fscale=0.8,
                hspacing=0.75,
                lspacing=0.15,
                dotsep="x",
                islocal=False,
            ),
            Vega=dict(
                mono=False,
                fscale=0.8,
                hspacing=0.75,
                lspacing=0.15,
                dotsep="×",
                islocal=False,
            ),
            Meson=dict(
                mono=False,
                fscale=0.8,
                hspacing=0.9,
                lspacing=0.225,
                dotsep="~×",
                islocal=False,
            ),
            Komika=dict(
                mono=False,
                fscale=0.7,
                hspacing=0.75,
                lspacing=0.225,
                dotsep="~×",
                islocal=False,
            ),
            Brachium=dict(
                mono=True,
                fscale=0.8,
                hspacing=1,
                lspacing=0.1,
                dotsep="x",
                islocal=False,
            ),
            Dalim=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=1,
                dotsep="~×",
                islocal=False,
            ),
            Miro=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=1,
                dotsep="~×",
                islocal=False,
            ),
            Ubuntu=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=1,
                dotsep="~×",
                islocal=False,
            ),
            Mizar=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=0.75,
                dotsep="~×",
                islocal=False,
            ),
            LiberationSans=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=1,
                dotsep="~×",
                islocal=False,
            ),
            DejavuSansMono=dict(
                mono=True,
                fscale=0.725,
                hspacing=1,
                lspacing=0.1,
                dotsep="~×",
                islocal=False,
            ),
            SunflowerHighway=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=1,
                dotsep="~×",
                islocal=False,
            ),
            Swansea=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=1,
                dotsep="~×",
                islocal=False,
            ),
            Housekeeper=dict(  # supports chinese glyphs
                mono=False,
                fscale=0.75,
                hspacing=1,
                lspacing=0.2,
                dotsep="~×",
                islocal=False,
            ),
            Wananti=dict(  # supports chinese glyphs
                mono=False,
                fscale=0.75,
                hspacing=1,
                lspacing=0.2,
                dotsep="~x",
                islocal=False,
            ),
            AnimeAce=dict(
                mono=False,
                fscale=0.75,
                hspacing=1,
                lspacing=0.2,
                dotsep="~x",
                islocal=False,
            ),
            Antares=dict(
                mono=True,
                fscale=0.8,
                hspacing=1,
                lspacing=0.1,
                dotsep="×",
                islocal=False,
            ),
            Archistico=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=0.75,
                dotsep="~×",
                islocal=False,
            ),
            KazyCase=dict(
                mono=True,
                fscale=0.8,
                hspacing=1.2,
                lspacing=0.1,
                dotsep="×",
                islocal=False,
            ),
            Roboto=dict(
                mono=True,
                fscale=0.8,
                hspacing=1,
                lspacing=0.1,
                dotsep="×",
                islocal=False,
            ),
        )

    ####################################################################################
    def __getitem__(self, key):
        """Make the class work like a dictionary too"""
        return getattr(self, key)

    def __setitem__(self, key, value):
        """Make the class work like a dictionary too"""
        setattr(self, key, value)

    def __str__(self) -> str:
        """Return a string representation of the object"""
        s = Settings.__doc__.replace("   ", "")
        s = s.replace(".. code-block:: python\n", "")
        s = s.replace("```python\n", "")
        s = s.replace("```\n", "")
        s = s.replace("\n\n", "\n #------------------------------------------------------\n")
        s = s.replace("\n  ", "\n")
        s = s.replace("\n ", "\n")
        s = s.replace(" from", "from")
        try:
            from pygments import highlight
            from pygments.lexers import Python3Lexer
            from pygments.formatters import Terminal256Formatter
            s = highlight(s, Python3Lexer(), Terminal256Formatter(style="zenburn"))
        except (ModuleNotFoundError, ImportError):
            pass

        module = self.__class__.__module__
        name = self.__class__.__name__
        header = f"{module}.{name} at ({hex(id(self))})".ljust(75)
        s = f"\x1b[1m\x1b[7m{header}\x1b[0m\n" + s
        return s.strip()

    ############################################################
    def keys(self) -> list:
        """Return all keys"""
        return self.__slots__

    def values(self) -> list:
        """Return all values"""
        return [getattr(self, key) for key in self.__slots__]

    def items(self) -> list:
        """Return all items"""
        return [(key, getattr(self, key)) for key in self.__slots__]

    def reset(self) -> None:
        """Reset all settings to their default status."""
        self.__init__()

    ############################################################
    def init_colab(self, enable_k3d=True) -> None:
        """
        Initialize colab environment
        """
        print("setting up colab environment (can take a minute) ...", end="")

        res = os.system("which Xvfb")
        if res:
            os.system("apt-get install xvfb")

        os.system("pip install pyvirtualdisplay")

        from pyvirtualdisplay import Display
        Display(visible=0).start()

        if enable_k3d:
            os.system("pip install k3d")

        from google.colab import output
        output.enable_custom_widget_manager()

        if enable_k3d:
            import k3d
            try:
                print("installing k3d...", end="")
                os.system("jupyter nbextension install --py --user k3d")
                os.system("jupyter nbextension enable  --py --user k3d")
                k3d.switch_to_text_protocol()
                self.default_backend = "k3d"
                self.backend_autoclose = False
            except:
                print("(FAILED) ... ", end="")

        print(" setup completed.")

    ############################################################
    def start_xvfb(self) -> None:
        """
        Start xvfb.

        Xvfb or X virtual framebuffer is a display server implementing
        the X11 display server protocol. In contrast to other display servers,
        Xvfb performs all graphical operations in virtual memory
        without showing any screen output.
        """
        print("starting xvfb (can take a minute) ...", end="")
        res = os.system("which Xvfb")
        if res:
            os.system("apt-get install xvfb")
        os.system("set -x")
        os.system("export DISPLAY=:99.0")
        os.system("Xvfb :99 -screen 0 1024x768x24 > /dev/null 2>&1 &")
        os.system("sleep 3")
        os.system("set +x")
        os.system('exec "$@"')
        print(" xvfb started.")



from vedo import *

settings.immediate_rendering = False  # faster for multi-renderers

# (0,0) is the bottom-left corner of the window, (1,1) the top-right
# the order in the list defines the priority when overlapping
custom_shape = [
    dict(bottomleft=(0.0, 0.0), topright=(0.5, 1), bg="wheat", bg2="w"),  # ren0
    dict(bottomleft=(0.5, 0.0), topright=(1, 1), bg="blue3", bg2="lb"),  # ren1
    dict(bottomleft=(0.2, 0.05), topright=(0.8, 0.1), bg="white"),  # ren2
]
plt = Plotter(shape=custom_shape, size=(1600, 900))

s0 = ParametricShape(0)
s1 = ParametricShape(1)

plt.at(0).add(s0, "renderer0")
plt.at(1).add(s1, "renderer1")


def slider1(widget, event):
    value = widget.value
    s0.rotate_y(value)
    s1.rotate_y(-value)

opts = dict(
    slider_length=0.06,
    slider_width=0.6,
    end_cap_length=0.02,
    end_cap_width=0.5,
    tube_width=0.1,
    title_height=0.15,
)
plt.at(2).add_slider(slider1, -5, 5, value=0, pos=([0.05, 0.02], [0.55, 0.02]), **opts)

plt.show(interactive=True).close()


"""Specify color mapping for cells
and points of a Mesh"""
from vedo import dataurl, Plotter, Mesh

plt = Plotter(N=3, axes=11)

##################################### add a cell array
man1 = Mesh(dataurl+"man_low.vtk").linewidth(0.1)
nv = man1.ncells                    # nr. of cells
scals = range(nv)                   # coloring by the index of cell
man1.cmap("Paired", scals, on='cells').add_scalarbar("cell nr")
plt.at(0).show(man1, __doc__, elevation=-60)


##################################### Point coloring
man2 = Mesh(dataurl+"man_low.vtk")
scals = man2.vertices[:, 0] + 37    # pick x coordinates of vertices
man2.cmap("hot", scals)
man2.add_scalarbar(horizontal=True)
plt.at(1).show(man2, "mesh.cmap()")


##################################### Cell coloring
man3 = Mesh(dataurl+"man_low.vtk")
cell_centers = man3.cell_centers().coordinates
scals = cell_centers[:, 2] + 37  # pick z coordinates of cells
man3.cmap("afmhot", scals, on='cells')

# Add a fancier 3D scalar bar embedded in the 3d scene
man3.add_scalarbar3d(size=[0.2,3])
man3.scalarbar.scale(1.1).rotate_x(90).shift([0,2,0])

plt.at(2).show(man3, "mesh.cmap(on='cells')")
plt.interactive()
plt.close()



import numpy as np
from vedo import settings, Plotter, Arrow, screenshot, \
    ScalarBar3D, Axes, mag, color_map, Assembly, Line

settings.default_font = 'Theemim'
settings.multi_samples=8

filename = 'data/obs_xyz.dat'
xyz = np.loadtxt(filename, dtype=np.float64)
xyz = xyz[1:, :] - np.array([0, 20, 0])
n_obs, n_col = xyz.shape

# Read in the data
num_field = np.loadtxt('data/tmp_field.txt', dtype=np.float64)
amp = mag(num_field)

# We need to create a few slices from the original arrays
# First create a mask array to mark all receiver points with x = -10, 0, 10
x_mark = [-19.5, -9.236842110, 1.026315790, 11.289473680, 19.5]
y_mark = [-4.5, 4.5]
mask = np.zeros(n_obs, dtype=bool)

# We need to create a mask array to mark all receiver points with x being
# any of x_mark and y being any of y_marko
for x_m in x_mark:
    mask_local = np.abs(xyz[:, 0] - x_m) < 1e-6
    mask = np.logical_or(mask, mask_local)

# Create an Arrows object for all the receivers where mask is True
start = xyz[mask, :]
orientation = num_field[mask, :]
orientation = orientation / mag(orientation)[:, None] # normalize
amp_mask = amp[mask]
vrange = np.array([amp_mask.min(), amp_mask.max()])

arrs = []
for i in range(start.shape[0]):
    arr = Arrow(start[i], start[i] + orientation[i] * 4)
    color = color_map(
        amp_mask[i], "jet", 
        vmin=vrange[0], vmax=vrange[1],
    )
    arr.color(color).lighting('off')
    arrs.append(arr)
arrows = Assembly(arrs)

# create a 2D scalarbar
# scalarbar = ScalarBar(
#     vrange,
#     title='E (V/m)',
#     c="jet",
#     font_size=22, 
#     pos=(0.7, 0.25),
#     size=(60,600),
# )

# create a dummy line to use as a 3D scalarbar
pos = (-10, -14, -32)
line = Line([0, 0], [1, 0]).cmap("jet", vrange * 1e6)
scalarbar = ScalarBar3D(
    line,
    # c="white",
    title="E (:muV/m)",
    title_size=3,
    label_rotation=90,
    label_offset=.5,
    label_size=2,
    pos=pos,
    size=(1, 20),
    nlabels=5,
)
scalarbar.rotate_z(-90)

size = (3920, 2160)
plt = Plotter()

axes = Axes(
    arrows, 
    xtitle='Easting (m)',
    ytitle='Northing (m)',
    ztitle='Elevation (m)',
    xtitle_position=0.60,
    xlabel_size=0.018,
    xtitle_offset=0.15,
    ytitle_position=0.85,
    ylabel_rotation=-90,
    ylabel_size=0.02,
    ytitle_rotation=180,
    y_values_and_labels=[(-5, "-5"), (0, "0"), (5, "5")],
    axes_linewidth=4,
    xrange=arrows.xbounds(),
    yrange=arrows.ybounds(),
    zxgrid2=True,
    zshift_along_y=1,
    zaxis_rotation=-70,
    ztitle_size=0.02,
    ztitle_position=0.68,
    xyframe_line=True,
    grid_linewidth=2,
)
cam = dict(
    position=(-58.8911, -54.5234, 8.63461),
    focal_point=(-5.25549, -0.0457020, -23.8989),
    viewup=(0.248841, 0.304150, 0.919549),
    distance=83.0844,
    clipping_range=(34.8493, 143.093),
)
fig_name = 'data/electric_field.png'
plt.show(arrows, axes, scalarbar, interactive=0, camera=cam)
# screenshot(fig_name)
plt.interactive().close()


"""Probe a Volume with a Mesh"""
from vedo import *

# Load a Volume
vol = Volume(dataurl + 'embryo.slc')
vol.cmap('bone').mode(1)

# Create a Mesh (can be any mesh)
msh = Paraboloid(res=200).scale(200).pos([100,100,200])

# Probe the Volume with the Mesh
# and colorize it with the probed values
msh.probe(vol)
msh.cmap('Spectral').add_scalarbar().print()

show(vol, msh, __doc__, axes=1).close()



"""Visualize stress tensors as ellipsoids"""
import vtk
from vedo import *

# Create a test volume with tensors
pl = vtk.vtkPointLoad()
pl.SetLoadValue(50)
pl.SetSampleDimensions(6,6,6)
pl.ComputeEffectiveStressOn()
pl.SetPoissonsRatio(0.2)
pl.SetModelBounds(-10,10,-10,10,-10,10)
pl.Update()

vol = Volume(pl.GetOutput()).mode(1)
print(vol)

# Extract a slice of the volume data at index 3
zsl = vol.zslice(3)

# Generate tensor ellipsoids
tens1 = Tensors(vol, source='ellipse', scale=10).cmap("Reds")
print(tens1)

tens2 = Tensors(zsl, source='ellipse', scale=20).cmap("Greens")
print(tens2)

show([(vol, __doc__), tens1], N=2, axes=9, viewup='z').close()

show(vol, tens2, zsl, axes=9, viewup='z').close()

